# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from collections.abc import MutableMapping
    from collections.abc import Sequence
except ImportError:
    from collections import MutableMapping
    from collections import Sequence

import threading
from collections import OrderedDict, deque
from datetime import datetime
from itertools import chain

from opencensus.common import utils
from opencensus.trace import attributes as attributes_module
from opencensus.trace import base_span
from opencensus.trace import link as link_module
from opencensus.trace import stack_trace as stack_trace_module
from opencensus.trace import status as status_module
from opencensus.trace import time_event
from opencensus.trace.span_context import generate_span_id
from opencensus.trace.tracers import base

# https://github.com/census-instrumentation/opencensus-specs/blob/master/trace/TraceConfig.md  # noqa
MAX_NUM_ATTRIBUTES = 32
MAX_NUM_ANNOTATIONS = 32
MAX_NUM_MESSAGE_EVENTS = 128
MAX_NUM_LINKS = 32


class BoundedList(Sequence):
    """An append only list with a fixed max size."""
    def __init__(self, maxlen):
        self.dropped = 0
        self._dq = deque(maxlen=maxlen)
        self._lock = threading.Lock()

    def __repr__(self):
        return ("{}({}, maxlen={})"
                .format(
                    type(self).__name__,
                    list(self._dq),
                    self._dq.maxlen
                ))

    def __getitem__(self, index):
        return self._dq[index]

    def __len__(self):
        return len(self._dq)

    def __iter__(self):
        return iter(self._dq)

    def append(self, item):
        with self._lock:
            if len(self._dq) == self._dq.maxlen:
                self.dropped += 1
            self._dq.append(item)

    def extend(self, seq):
        with self._lock:
            to_drop = len(seq) + len(self._dq) - self._dq.maxlen
            if to_drop > 0:
                self.dropped += to_drop
            self._dq.extend(seq)

    @classmethod
    def from_seq(cls, maxlen, seq):
        seq = tuple(seq)
        if len(seq) > maxlen:
            raise ValueError
        bounded_list = cls(maxlen)
        bounded_list._dq = deque(seq, maxlen=maxlen)
        return bounded_list


class BoundedDict(MutableMapping):
    """A dict with a fixed max capacity."""
    def __init__(self, maxlen):
        self.maxlen = maxlen
        self.dropped = 0
        self._dict = OrderedDict()
        self._lock = threading.Lock()

    def __repr__(self):
        return ("{}({}, maxlen={})"
                .format(
                    type(self).__name__,
                    dict(self._dict),
                    self.maxlen
                ))

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        with self._lock:
            if key in self._dict:
                del self._dict[key]
            elif len(self._dict) == self.maxlen:
                del self._dict[next(iter(self._dict.keys()))]
                self.dropped += 1
            self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    @classmethod
    def from_map(cls, maxlen, mapping):
        mapping = OrderedDict(mapping)
        if len(mapping) > maxlen:
            raise ValueError
        bounded_dict = cls(maxlen)
        bounded_dict._dict = mapping
        return bounded_dict


class SpanKind(object):
    UNSPECIFIED = 0
    SERVER = 1
    CLIENT = 2


class Span(base_span.BaseSpan):
    """A span is an individual timed event which forms a node of the trace
    tree. Each span has its name, span id and parent id. The parent id
    indicates the causal relationships between the individual spans in a
    single distributed trace. Span that does not have a parent id is called
    root span. All spans associated with a specific trace also share a common
    trace id. Spans do not need to be continuous, there can be gaps between
    two spans.

    :type name: str
    :param name: The name of the span.

    :type parent_span: :class:`~opencensus.trace.span.Span`
    :param parent_span: (Optional) Parent span.

    :type attributes: dict
    :param attributes: Collection of attributes associated with the span.
                   Attribute keys must be less than 128 bytes.
                   Attribute values must be less than 16 kilobytes.

    :type start_time: str
    :param start_time: (Optional) Start of the time interval (inclusive)
                       during which the trace data was collected from the
                       application.

    :type end_time: str
    :param end_time: (Optional) End of the time interval (inclusive) during
                     which the trace data was collected from the application.

    :type span_id: int
    :param span_id: Identifier for the span, unique within a trace.

    :type stack_trace: :class: `~opencensus.trace.stack_trace.StackTrace`
    :param stack_trace: (Optional) A call stack appearing in a trace

    :type annotations: list(:class:`opencensus.trace.time_event.Annotation`)
    :param annotations: (Optional) The list of span annotations.

    :type message_events:
        list(:class:`opencensus.trace.time_event.MessageEvent`)
    :param message_events: (Optional) The list of span message events.

    :type links: list
    :param links: (Optional) Links associated with the span. You can have up
                  to 128 links per Span.

    :type status: :class: `~opencensus.trace.status.Status`
    :param status: (Optional) An optional final status for this span.

    :type same_process_as_parent_span: bool
    :param same_process_as_parent_span: (Optional) A highly recommended but not
                                        required flag that identifies when a
                                        trace crosses a process boundary.
                                        True when the parent_span belongs to
                                        the same process as the current span.

    :type context_tracer: :class:`~opencensus.trace.tracers.context_tracer.
                                 ContextTracer`
    :param context_tracer: The tracer that holds a stack of spans. If this is
                           not None, then when exiting a span, use the end_span
                           method in the tracer class to finish a span. If no
                           tracer is passed in, then just finish the span using
                           the finish method in the Span class.

    :type span_kind: int
    :param span_kind: (Optional) Highly recommended flag that denotes the type
                        of span (valid values defined by :class:
                        `opencensus.trace.span.SpanKind`)
    """

    def __init__(
            self,
            name,
            parent_span=None,
            attributes=None,
            start_time=None,
            end_time=None,
            span_id=None,
            stack_trace=None,
            annotations=None,
            message_events=None,
            links=None,
            status=None,
            same_process_as_parent_span=None,
            context_tracer=None,
            span_kind=SpanKind.UNSPECIFIED):
        self.name = name
        self.parent_span = parent_span
        self.start_time = start_time
        self.end_time = end_time

        if span_id is None:
            span_id = generate_span_id()

        if attributes is None:
            self.attributes = BoundedDict(MAX_NUM_ATTRIBUTES)
        else:
            self.attributes = BoundedDict.from_map(
                MAX_NUM_ATTRIBUTES, attributes)

        # Do not manipulate spans directly using the methods in Span Class,
        # make sure to use the Tracer.
        if parent_span is None:
            parent_span = base.NullContextManager()

        if annotations is None:
            self.annotations = BoundedList(MAX_NUM_ANNOTATIONS)
        else:
            self.annotations = BoundedList.from_seq(MAX_NUM_LINKS, annotations)

        if message_events is None:
            self.message_events = BoundedList(MAX_NUM_MESSAGE_EVENTS)
        else:
            self.message_events = BoundedList.from_seq(
                MAX_NUM_LINKS, message_events)

        if links is None:
            self.links = BoundedList(MAX_NUM_LINKS)
        else:
            self.links = BoundedList.from_seq(MAX_NUM_LINKS, links)

        if status is None:
            self.status = status_module.Status.as_ok()
        else:
            self.status = status

        self.span_id = span_id
        self.stack_trace = stack_trace
        self.same_process_as_parent_span = same_process_as_parent_span
        self._child_spans = []
        self.context_tracer = context_tracer
        self.span_kind = span_kind
        for callback in Span._on_create_callbacks:
            callback(self)

    _on_create_callbacks = []

    @staticmethod
    def on_create(callback):
        Span._on_create_callbacks.append(callback)

    @property
    def children(self):
        """The child spans of the current span."""
        return self._child_spans

    def span(self, name='child_span'):
        """Create a child span for the current span and append it to the child
        spans list.

        :type name: str
        :param name: (Optional) The name of the child span.

        :rtype: :class: `~opencensus.trace.span.Span`
        :returns: A child Span to be added to the current span.
        """
        child_span = Span(name, parent_span=self)
        self._child_spans.append(child_span)
        return child_span

    def add_attribute(self, attribute_key, attribute_value):
        """Add attribute to span.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        self.attributes[attribute_key] = attribute_value

    def add_annotation(self, description, **attrs):
        """Add an annotation to span.

        :type description: str
        :param description: A user-supplied message describing the event.
                        The maximum length for the description is 256 bytes.

        :type attrs: kwargs
        :param attrs: keyworded arguments e.g. failed=True, name='Caching'
        """
        self.annotations.append(time_event.Annotation(
            datetime.utcnow(),
            description,
            attributes_module.Attributes(attrs)
        ))

    def add_message_event(self, message_event):
        """Add a message event to this span.

        :type message_event: :class:`opencensus.trace.time_event.MessageEvent`
        :param message_event: The message event to attach to this span.
        """
        self.message_events.append(message_event)

    def add_link(self, link):
        """Add a Link.

        :type link: :class: `~opencensus.trace.link.Link`
        :param link: A Link object.
        """
        if isinstance(link, link_module.Link):
            self.links.append(link)
        else:
            raise TypeError("Type Error: received {}, but requires Link.".
                            format(type(link).__name__))

    def set_status(self, status):
        """Sets span status.

        :type code: :class: `~opencensus.trace.status.Status`
        :param code: A Status object.
        """
        if isinstance(status, status_module.Status):
            self.status = status
        else:
            raise TypeError("Type Error: received {}, but requires Status.".
                            format(type(status).__name__))

    def start(self):
        """Set the start time for a span."""
        self.start_time = utils.to_iso_str()

    def finish(self):
        """Set the end time for a span."""
        self.end_time = utils.to_iso_str()

    def __iter__(self):
        """Iterate through the span tree."""
        for span in chain.from_iterable(map(iter, self.children)):
            yield span
        yield self

    def __enter__(self):
        """Start a span."""
        self.start()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Finish a span."""
        if traceback is not None:
            self.stack_trace =\
                stack_trace_module.StackTrace.from_traceback(traceback)
        if exception_value is not None:
            self.status = status_module.Status.from_exception(exception_value)
        if self.context_tracer is not None:
            self.context_tracer.end_span()
            return

        self.finish()


def format_span_json(span):
    """Helper to format a Span in JSON format.

    :type span: :class:`~opencensus.trace.span.Span`
    :param span: A Span to be transferred to JSON format.

    :rtype: dict
    :returns: Formatted Span.
    """
    span_json = {
        'displayName': utils.get_truncatable_str(span.name),
        'spanId': span.span_id,
        'startTime': span.start_time,
        'endTime': span.end_time,
        'childSpanCount': len(span._child_spans)
    }

    parent_span_id = None

    if span.parent_span is not None:
        parent_span_id = span.parent_span.span_id

    if parent_span_id is not None:
        span_json['parentSpanId'] = parent_span_id

    if span.attributes:
        span_json['attributes'] = attributes_module.Attributes(
            span.attributes).format_attributes_json()

    if span.stack_trace is not None:
        span_json['stackTrace'] = span.stack_trace.format_stack_trace_json()

    formatted_time_events = []
    if span.annotations:
        formatted_time_events.extend(
            {'time': aa.timestamp,
             'annotation': aa.format_annotation_json()}
            for aa in span.annotations)
    if span.message_events:
        formatted_time_events.extend(
            {'time': aa.timestamp,
             'message_event': aa.format_message_event_json()}
            for aa in span.message_events)
    if formatted_time_events:
        span_json['timeEvents'] = {
            'timeEvent': formatted_time_events
        }

    if span.links:
        span_json['links'] = {
            'link': [
                link.format_link_json() for link in span.links]
        }

    if span.status is not None:
        span_json['status'] = span.status.format_status_json()

    if span.same_process_as_parent_span is not None:
        span_json['sameProcessAsParentSpan'] = \
            span.same_process_as_parent_span

    return span_json



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

import threading
from collections import OrderedDict
from datetime import datetime

from opencensus.common import utils
from opencensus.metrics.export import (
    metric,
    metric_descriptor,
    metric_producer,
)
from opencensus.metrics.export import point as point_module
from opencensus.metrics.export import time_series
from opencensus.metrics.export import value as value_module


def get_timeseries_list(points, timestamp):
    """Convert a list of `GaugePoint`s into a list of `TimeSeries`.

    Get a :class:`opencensus.metrics.export.time_series.TimeSeries` for each
    measurement in `points`. Each series contains a single
    :class:`opencensus.metrics.export.point.Point` that represents the last
    recorded value of the measurement.

    :type points: list(:class:`GaugePoint`)
    :param points: The list of measurements to convert.

    :type timestamp: :class:`datetime.datetime`
    :param timestamp: Recording time to report, usually the current time.

    :rtype: list(:class:`opencensus.metrics.export.time_series.TimeSeries`)
    :return: A list of one `TimeSeries` for each point in `points`.
    """
    ts_list = []
    for lv, gp in points.items():
        point = point_module.Point(gp.to_point_value(), timestamp)
        ts_list.append(time_series.TimeSeries(lv, [point], timestamp))
    return ts_list


class GaugePoint(object):

    def to_point_value(self):
        raise NotImplementedError  # pragma: NO COVER

    def get_value(self):
        raise NotImplementedError  # pragma: NO COVER


class GaugePointLong(GaugePoint):
    """An instantaneous measurement from a LongGauge.

    A GaugePointLong represents the most recent measurement from a
    :class:`LongGauge` for a given set of label values.
    """

    def __init__(self):
        self.value = 0
        self._value_lock = threading.Lock()

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value
                ))

    def add(self, val):
        """Add `val` to the current value.

        :type val: int
        :param val: Value to add.
        """
        if not isinstance(val, six.integer_types):
            raise ValueError("GaugePointLong only supports integer types")
        with self._value_lock:
            self.value += val

    def _set(self, val):
        if not isinstance(val, six.integer_types):
            raise ValueError("GaugePointLong only supports integer types")
        with self._value_lock:
            self.value = val

    def set(self, val):
        """Set the current value to `val`.

        :type val: int
        :param val: Value to set.
        """
        self._set(val)

    def get_value(self):
        """Get the current value.

        :rtype: int
        :return: The current value of the measurement.
        """
        return self.value

    def to_point_value(self):
        """Get a point value conversion of the current value.

        :rtype: :class:`opencensus.metrics.export.value.ValueLong`
        :return: A converted `ValueLong`.
        """
        return value_module.ValueLong(self.value)


class GaugePointDouble(GaugePoint):
    """An instantaneous measurement from a DoubleGauge.

    A `GaugePointDouble` represents the most recent measurement from a
    :class:`DoubleGauge` for a given set of label values.
    """

    def __init__(self):
        self.value = 0.0
        self._value_lock = threading.Lock()

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value
                ))

    def add(self, val):
        """Add `val` to the current value.

        :type val: float
        :param val: Value to add.
        """
        with self._value_lock:
            self.value += val

    def _set(self, val):
        with self._value_lock:
            self.value = float(val)

    def set(self, val):
        """Set the current value to `val`.

        :type val: float
        :param val: Value to set.
        """
        self._set(val)

    def get_value(self):
        """Get the current value.

        :rtype: float
        :return: The current value of the measurement.
        """
        return self.value

    def to_point_value(self):
        """Get a point value conversion of the current value.

        :rtype: :class:`opencensus.metrics.export.value.ValueDouble`
        :return: A converted `ValueDouble`.
        """
        return value_module.ValueDouble(self.value)


class DerivedGaugePoint(GaugePoint):
    """Wraps a `GaugePoint` to automatically track the value of a function.

    A `DerivedGaugePoint` is a read-only measure that stores the most recently
    read value of a given function in a mutable `GaugePoint`. Calling
    `get_value` or `to_point_value` calls the tracked function and updates the
    wrapped `GaugePoint`.

    :type func: function
    :param func: The function to track.

    :type gauge_point: :class:`GaugePointLong`, :class:`GaugePointDouble`,
        :class:`opencensus.metrics.export.cumulative.CumulativePointLong`, or
        :class:`opencensus.metrics.export.cumulative.CumulativePointDouble`
    :param gauge_point: The underlying `GaugePoint`.
    """
    def __init__(self, func, gauge_point, **kwargs):
        self.gauge_point = gauge_point
        self.func = utils.get_weakref(func)
        self._kwargs = kwargs

    def __repr__(self):
        return ("{}({})({})"
                .format(
                    type(self).__name__,
                    self.func(),
                    self._kwargs
                ))

    def get_value(self):
        """Get the current value of the underlying measurement.

        Calls the tracked function and stores the value in the wrapped
        measurement as a side-effect.

        :rtype: int, float, or None
        :return: The current value of the wrapped function, or `None` if it no
            longer exists.
        """
        try:
            val = self.func()(**self._kwargs)
        except TypeError:  # The underlying function has been GC'd
            return None

        self.gauge_point._set(val)
        return self.gauge_point.get_value()

    def to_point_value(self):
        """Get a point value conversion of the current value.

        Calls the tracked function and stores the value in the wrapped
        measurement as a side-effect.

        :rtype: :class:`opencensus.metrics.export.value.ValueLong`,
            :class:`opencensus.metrics.export.value.ValueDouble`, or None
        :return: The point value conversion of the underlying `GaugePoint`, or
            None if the tracked function no longer exists.
        """
        if self.get_value() is None:
            return None
        return self.gauge_point.to_point_value()


class BaseGauge(object):
    """Base class for sets instantaneous measurements."""

    def __init__(self, name, description, unit, label_keys):
        self._len_label_keys = len(label_keys)
        self.default_label_values = [None] * self._len_label_keys
        self.descriptor = metric_descriptor.MetricDescriptor(
            name, description, unit, self.descriptor_type, label_keys)
        self.points = OrderedDict()
        self._points_lock = threading.Lock()

    def __repr__(self):
        return ('{}(descriptor.name="{}", points={})'
                .format(
                    type(self).__name__,
                    self.descriptor.name,
                    self.points
                ))

    def _remove_time_series(self, label_values):
        with self._points_lock:
            try:
                del self.points[tuple(label_values)]
            except KeyError:
                pass

    def remove_time_series(self, label_values):
        """Remove the time series for specific label values.

        :type label_values: list(:class:`LabelValue`)
        :param label_values: Label values of the time series to remove.
        """
        if label_values is None:
            raise ValueError
        if any(lv is None for lv in label_values):
            raise ValueError
        if len(label_values) != self._len_label_keys:
            raise ValueError
        self._remove_time_series(label_values)

    def remove_default_time_series(self):
        """Remove the default time series for this gauge."""
        self._remove_time_series(self.default_label_values)

    def clear(self):
        """Remove all points from this gauge."""
        with self._points_lock:
            self.points = OrderedDict()

    def get_metric(self, timestamp):
        """Get a metric including all current time series.

        Get a :class:`opencensus.metrics.export.metric.Metric` with one
        :class:`opencensus.metrics.export.time_series.TimeSeries` for each
        set of label values with a recorded measurement. Each `TimeSeries`
        has a single point that represents the last recorded value.

        :type timestamp: :class:`datetime.datetime`
        :param timestamp: Recording time to report, usually the current time.

        :rtype: :class:`opencensus.metrics.export.metric.Metric` or None
        :return: A converted metric for all current measurements.
        """
        if not self.points:
            return None

        with self._points_lock:
            ts_list = get_timeseries_list(self.points, timestamp)
        return metric.Metric(self.descriptor, ts_list)

    @property
    def descriptor_type(self):  # pragma: NO COVER
        raise NotImplementedError

    @property
    def point_type(self):  # pragma: NO COVER
        raise NotImplementedError


class Gauge(BaseGauge):
    """A set of mutable, instantaneous measurements of the same type.

    End users should use :class:`LongGauge`, :class:`DoubleGauge`,
    :class:`opencensus.metrics.export.cumulative.LongCumulative`, or
    :class:`opencensus.metrics.export.cumulative.DoubleCumulative` instead of
    using this class directly.

    The constructor arguments are used to create a
    :class:`opencensus.metrics.export.metric_descriptor.MetricDescriptor` for
    converted metrics. See that class for details.
    """

    def _get_or_create_time_series(self, label_values):
        with self._points_lock:
            return self.points.setdefault(
                tuple(label_values), self.point_type())

    def get_or_create_time_series(self, label_values):
        """Get a mutable measurement for the given set of label values.

        :type label_values: list(:class:`LabelValue`)
        :param label_values: The measurement's label values.

        :rtype: :class:`GaugePointLong`, :class:`GaugePointDouble`
            :class:`opencensus.metrics.export.cumulative.CumulativePointLong`,
            or
            :class:`opencensus.metrics.export.cumulative.CumulativePointDouble`
        :return: A mutable point that represents the last value of the
            measurement.
        """
        if label_values is None:
            raise ValueError
        if any(lv is None for lv in label_values):
            raise ValueError
        if len(label_values) != self._len_label_keys:
            raise ValueError
        return self._get_or_create_time_series(label_values)

    def get_or_create_default_time_series(self):
        """Get the default measurement for this gauge.

        Each gauge has a default point not associated with any specific label
        values. When this gauge is exported as a metric via `get_metric` the
        time series associated with this point will have null label values.

        :rtype: :class:`GaugePointLong`, :class:`GaugePointDouble`
            :class:`opencensus.metrics.export.cumulative.CumulativePointLong`,
            or
            :class:`opencensus.metrics.export.cumulative.CumulativePointDouble`
        :return: A mutable point that represents the last value of the
            measurement.
        """
        return self._get_or_create_time_series(self.default_label_values)


class LongGaugeMixin(object):
    """Type mixin for long-valued gauges."""
    descriptor_type = metric_descriptor.MetricDescriptorType.GAUGE_INT64
    point_type = GaugePointLong


class DoubleGaugeMixin(object):
    """Type mixin for float-valued gauges."""
    descriptor_type = metric_descriptor.MetricDescriptorType.GAUGE_DOUBLE
    point_type = GaugePointDouble


class LongGauge(LongGaugeMixin, Gauge):
    """Gauge for recording int-valued measurements."""


class DoubleGauge(DoubleGaugeMixin, Gauge):
    """Gauge for recording float-valued measurements."""


class DerivedGauge(BaseGauge):
    """Gauge that tracks values of other functions.

    Each of a `DerivedGauge`'s measurements are associated with a function
    which is called when the gauge is exported.

    End users should use :class:`DerivedLongGauge`, :class:`DerivedDoubleGauge`
    :class:`opencensus.metrics.export.cumulative.DerivedLongCumulative`, or
    :class:`opencensus.metrics.export.cumulative.DerivedDoubleCumulative`
    instead of using this class directly.
    """

    def _create_time_series(self, label_values, func, **kwargs):
        with self._points_lock:
            return self.points.setdefault(
                tuple(label_values),
                DerivedGaugePoint(func, self.point_type(), **kwargs))

    def create_time_series(self, label_values, func, **kwargs):
        """Create a derived measurement to trac `func`.

        :type label_values: list(:class:`LabelValue`)
        :param label_values: The measurement's label values.

        :type func: function
        :param func: The function to track.

        :rtype: :class:`DerivedGaugePoint`
        :return: A read-only measurement that tracks `func`.
        """
        if label_values is None:
            raise ValueError
        if any(lv is None for lv in label_values):
            raise ValueError
        if len(label_values) != self._len_label_keys:
            raise ValueError
        if func is None:
            raise ValueError
        return self._create_time_series(label_values, func, **kwargs)

    def create_default_time_series(self, func):
        """Create the default derived measurement for this gauge.

        :type func: function
        :param func: The function to track.

        :rtype: :class:`DerivedGaugePoint`
        :return: A read-only measurement that tracks `func`.
        """
        if func is None:
            raise ValueError
        return self._create_time_series(self.default_label_values, func)


class DerivedLongGauge(LongGaugeMixin, DerivedGauge):
    """Gauge for derived int-valued measurements."""


class DerivedDoubleGauge(DoubleGaugeMixin, DerivedGauge):
    """Gauge for derived float-valued measurements."""


class Registry(metric_producer.MetricProducer):
    """A collection of gauges to be exported together.

    Each registered gauge must have a unique `descriptor.name`.
    """

    def __init__(self):
        self.gauges = {}
        self._gauges_lock = threading.Lock()

    def __repr__(self):
        return ('{}(gauges={}'
                .format(
                    type(self).__name__,
                    self.gauges
                ))

    def add_gauge(self, gauge):
        """Add `gauge` to the registry.

        Raises a `ValueError` if another gauge with the same name already
        exists in the registry.

        :type gauge: class:`LongGauge`, class:`DoubleGauge`,
            :class:`opencensus.metrics.export.cumulative.LongCumulative`,
            :class:`opencensus.metrics.export.cumulative.DoubleCumulative`,
            :class:`DerivedLongGauge`, :class:`DerivedDoubleGauge`
            :class:`opencensus.metrics.export.cumulative.DerivedLongCumulative`,
            or
            :class:`opencensus.metrics.export.cumulative.DerivedDoubleCumulative`
        :param gauge: The gauge to add to the registry.
        """
        if gauge is None:
            raise ValueError
        name = gauge.descriptor.name
        with self._gauges_lock:
            if name in self.gauges:
                raise ValueError(
                    'Another gauge named "{}" is already registered'
                    .format(name))
            self.gauges[name] = gauge

    def get_metrics(self):
        """Get a metric for each gauge in the registry at the current time.

        :rtype: set(:class:`opencensus.metrics.export.metric.Metric`)
        :return: A set of `Metric`s, one for each registered gauge.
        """
        now = datetime.utcnow()
        metrics = set()
        for gauge in self.gauges.values():
            metrics.add(gauge.get_metric(now))
        return metrics



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class LabelKey(object):
    """The label keys associated with the metric descriptor.

    :type key: str
    :param key: the key for the label

    :type description: str
    :param description: description of the label
    """
    def __init__(self, key, description):
        self._key = key
        self._description = description

    def __repr__(self):
        if self.description:
            return ('{}({}, description="{}")'
                    .format(
                        type(self).__name__,
                        self.key,
                        self.description
                    ))
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.key,
                ))

    @property
    def key(self):
        """the key for the label"""
        return self._key

    @property
    def description(self):
        """a human-readable description of what this label key represents"""
        return self._description



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # For Python 3.0 and later
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError, URLError
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen, Request
    from urllib2 import HTTPError, URLError


import socket

_REQUEST_TIMEOUT = 2  # in secs


def get_request(request_url, request_headers=dict()):
    """Execute http get request on given request_url with optional headers
    """
    request = Request(request_url)
    for key, val in request_headers.items():
        request.add_header(key, val)

    try:
        response = urlopen(request, timeout=_REQUEST_TIMEOUT)
        response_content = response.read()
    except (HTTPError, URLError, socket.timeout):
        response_content = None

    return response_content



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing base class for exporters."""


class StatsExporter(object):
    """Base class for opencensus stats exporters.

    Subclasses of :class:`Exporter` must override :meth:`export`.
    """

    def on_register_view(self, view):
        """
        :type view: object of :class:
            `~opencensus.stats.view.View`
        :param object of opencensus.stats.view.View view:
            View object to register
        """
        raise NotImplementedError  # pragma: NO COVER

    def emit(self, view_datas):
        """Send view and measurement to exporter record method,
        and then it will record on its own way.

        :type view_datas: object of :class:
            `~opencensus.stats.view_data.ViewData`
        :param list of opencensus.stats.view_data.ViewData ViewData:
            list of ViewData object to send to Stackdriver Monitoring
        """
        raise NotImplementedError  # pragma: NO COVER



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


DEFAULT_SAMPLING_RATE = 1e-4


class Sampler(object):
    """Base class for opencensus trace request samplers.

    Subclasses must override :meth:`should_sample`.
    """

    def should_sample(self, span_context):
        """Whether to sample this request.

        :type span_context: :class:`opencensus.trace.span_context.SpanContext`
        :param span_context: The span context.

        :rtype: bool
        :returns: Whether to sample the request according to the context.
        """
        raise NotImplementedError


class AlwaysOnSampler(Sampler):
    """Sampler that samples every request, regardless of trace options."""

    def should_sample(self, span_context):
        return True


class AlwaysOffSampler(Sampler):
    """Sampler that doesn't sample any request, regardless of trace options."""

    def should_sample(self, span_context):
        return False


class ProbabilitySampler(Sampler):
    """Sample a request at a fixed rate.

    :type rate: float
    :param rate: The rate of sampling.
    """
    def __init__(self, rate=None):
        if rate is None:
            rate = DEFAULT_SAMPLING_RATE

        if not 0 <= rate <= 1:
            raise ValueError('Rate must between 0 and 1.')

        self.rate = rate

    def should_sample(self, span_context):
        """Make the sampling decision based on the lower 8 bytes of the trace
        ID. If the value is less than the bound, return True, else False.

        :type span_context: :class:`opencensus.trace.span_context.SpanContext`
        :param span_context: The span context.

        :rtype: bool
        :returns: Whether to sample the request according to the context.
        """
        if span_context.trace_options.get_enabled():
            return True

        lower_long = get_lower_long_from_trace_id(span_context.trace_id)
        bound = self.rate * 0xffffffffffffffff
        return lower_long <= bound


def get_lower_long_from_trace_id(trace_id):
    """Returns the lower 8 bytes of the trace ID as a long value, assuming
    little endian order.

    :rtype: long
    :returns: Lower 8 bytes of trace ID
    """
    lower_bytes = trace_id[16:]
    lower_long = int(lower_bytes, 16)

    return lower_long



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from opencensus.trace.span_context import INVALID_SPAN_ID, SpanContext
from opencensus.trace.trace_options import TraceOptions

_STATE_HEADER_KEY = 'b3'
_TRACE_ID_KEY = 'x-b3-traceid'
_SPAN_ID_KEY = 'x-b3-spanid'
_SAMPLED_KEY = 'x-b3-sampled'


class B3FormatPropagator(object):
    """Propagator for the B3 HTTP header format.

    See: https://github.com/openzipkin/b3-propagation
    """

    def from_headers(self, headers):
        """Generate a SpanContext object from B3 propagation headers.

        :type headers: dict
        :param headers: HTTP request headers.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from B3 propagation headers.
        """
        if headers is None:
            return SpanContext(from_header=False)

        trace_id, span_id, sampled = None, None, None

        state = headers.get(_STATE_HEADER_KEY)
        if state:
            fields = state.split('-', 4)

            if len(fields) == 1:
                sampled = fields[0]
            elif len(fields) == 2:
                trace_id, span_id = fields
            elif len(fields) == 3:
                trace_id, span_id, sampled = fields
            elif len(fields) == 4:
                trace_id, span_id, sampled, _parent_span_id = fields
            else:
                return SpanContext(from_header=False)
        else:
            trace_id = headers.get(_TRACE_ID_KEY)
            span_id = headers.get(_SPAN_ID_KEY)
            sampled = headers.get(_SAMPLED_KEY)

        if sampled is not None:
            # The specification encodes an enabled tracing decision as "1".
            # In the wild pre-standard implementations might still send "true".
            # "d" is set in the single header case when debugging is enabled.
            sampled = sampled.lower() in ('1', 'd', 'true')
        else:
            # If there's no incoming sampling decision, it was deferred to us.
            # Even though we set it to False here, we might still sample
            # depending on the tracer configuration.
            sampled = False

        trace_options = TraceOptions()
        trace_options.set_enabled(sampled)

        # TraceId and SpanId headers both have to exist
        if not trace_id or not span_id:
            return SpanContext(trace_options=trace_options)

        # Convert 64-bit trace ids to 128-bit
        if len(trace_id) == 16:
            trace_id = '0'*16 + trace_id

        span_context = SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            trace_options=trace_options,
            from_header=True
        )

        return span_context

    def to_headers(self, span_context):
        """Convert a SpanContext object to B3 propagation headers.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :rtype: dict
        :returns: B3 propagation headers.
        """

        if not span_context.span_id:
            span_id = INVALID_SPAN_ID
        else:
            span_id = span_context.span_id

        sampled = span_context.trace_options.enabled

        return {
            _TRACE_ID_KEY: span_context.trace_id,
            _SPAN_ID_KEY: span_id,
            _SAMPLED_KEY: '1' if sampled else '0'
        }



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class LabelValue(object):
    """The label values associated with the TimeSeries.

    :type value: str
    :param value: the value for the label
    """
    def __init__(self, value=None):
        self._value = value

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    @property
    def value(self):
        """the value for the label"""
        return self._value

    def __eq__(self, other):
        return isinstance(other, LabelValue) and \
            self.value == other.value

    def __hash__(self):
        return hash(self.value)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from opencensus.metrics.export.metric_descriptor import MetricDescriptorType
from opencensus.stats import aggregation_data
from opencensus.stats import measure as measure_module

logger = logging.getLogger(__name__)


class SumAggregation(object):
    """Sum Aggregation describes that data collected and aggregated with this
    method will be summed

    :type sum: int or float
    :param sum: the initial sum to be used in the aggregation

    """
    def __init__(self, sum=None):
        self._initial_sum = sum or 0

    def new_aggregation_data(self, measure):
        """Get a new AggregationData for this aggregation."""
        value_type = MetricDescriptorType.to_type_class(
            self.get_metric_type(measure))
        return aggregation_data.SumAggregationData(
            value_type=value_type, sum_data=self._initial_sum)

    @staticmethod
    def get_metric_type(measure):
        """Get the MetricDescriptorType for the metric produced by this
        aggregation and measure.
        """
        if isinstance(measure, measure_module.MeasureInt):
            return MetricDescriptorType.CUMULATIVE_INT64
        if isinstance(measure, measure_module.MeasureFloat):
            return MetricDescriptorType.CUMULATIVE_DOUBLE
        raise ValueError


class CountAggregation(object):
    """Describes that the data collected and aggregated with this method will
    be turned into a count value

    :type count: int
    :param count: the initial count to be used in the aggregation

    """
    def __init__(self, count=0):
        self._initial_count = count

    def new_aggregation_data(self, measure=None):
        """Get a new AggregationData for this aggregation."""
        return aggregation_data.CountAggregationData(self._initial_count)

    @staticmethod
    def get_metric_type(measure):
        """Get the MetricDescriptorType for the metric produced by this
        aggregation and measure.
        """
        return MetricDescriptorType.CUMULATIVE_INT64


class DistributionAggregation(object):
    """Distribution Aggregation indicates that the desired aggregation is a
    histogram distribution

    :type boundaries: list(:class:'~opencensus.stats.bucket_boundaries.
                            BucketBoundaries')
    :param boundaries: the bucket endpoints

    """

    def __init__(self, boundaries=None):
        if boundaries:
            if not all(boundaries[ii] < boundaries[ii + 1]
                       for ii in range(len(boundaries) - 1)):
                raise ValueError("bounds must be sorted in increasing order")
            for ii, bb in enumerate(boundaries):
                if bb > 0:
                    break
            else:
                ii += 1
            if ii:
                logger.warning("Dropping %s non-positive bucket boundaries",
                               ii)
            boundaries = boundaries[ii:]

        self._boundaries = boundaries

    def new_aggregation_data(self, measure=None):
        """Get a new AggregationData for this aggregation."""
        return aggregation_data.DistributionAggregationData(
            0, 0, 0, None, self._boundaries)

    @staticmethod
    def get_metric_type(measure):
        """Get the MetricDescriptorType for the metric produced by this
        aggregation and measure.
        """
        return MetricDescriptorType.CUMULATIVE_DISTRIBUTION


class LastValueAggregation(object):
    """Describes that the data collected with this method will
    overwrite the last recorded value

    :type value: long
    :param count: the initial value to be used in the aggregation

    """
    def __init__(self, value=0):
        self._initial_value = value

    def new_aggregation_data(self, measure):
        """Get a new AggregationData for this aggregation."""
        value_type = MetricDescriptorType.to_type_class(
            self.get_metric_type(measure))
        return aggregation_data.LastValueAggregationData(
            value=self._initial_value, value_type=value_type)

    @staticmethod
    def get_metric_type(measure):
        """Get the MetricDescriptorType for the metric produced by this
        aggregation and measure.
        """
        if isinstance(measure, measure_module.MeasureInt):
            return MetricDescriptorType.GAUGE_INT64
        if isinstance(measure, measure_module.MeasureFloat):
            return MetricDescriptorType.GAUGE_DOUBLE
        raise ValueError



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Measurement(object):
    """ A measurement is an object with a measure and a value attached to it

    :type measure: :class: '~opencensus.stats.measure.Measure'
    :param measure: A measure to pass into the measurement

    :type value: int or float
    :param value: value of the measurement

    """
    def __init__(self, measure, value):
        self._measure = measure
        self._value = value

    @property
    def value(self):
        """The value of the current measurement"""
        return self._value

    @property
    def measure(self):
        """The measure of the current measurement"""
        return self._measure


class MeasurementInt(Measurement):
    """ Creates a new Integer Measurement """
    def __init__(self, measure, value):
        super(MeasurementInt, self).__init__(measure, value)


class MeasurementFloat(Measurement):
    """ Creates a new Float Measurement """
    def __init__(self, measure, value):
        super(MeasurementFloat, self).__init__(measure, value)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from opencensus.trace.propagation.tracestate_string_format import (
    TracestateStringFormatter,
)
from opencensus.trace.span_context import SpanContext
from opencensus.trace.trace_options import TraceOptions

_TRACEPARENT_HEADER_NAME = 'traceparent'
_TRACESTATE_HEADER_NAME = 'tracestate'
_TRACEPARENT_HEADER_FORMAT = \
    '^[ \t]*([0-9a-f]{2})-([0-9a-f]{32})-([0-9a-f]{16})-([0-9a-f]{2})' + \
    '(-.*)?[ \t]*$'
_TRACEPARENT_HEADER_FORMAT_RE = re.compile(_TRACEPARENT_HEADER_FORMAT)


class TraceContextPropagator(object):
    """Propagator for processing the trace context HTTP header format."""

    def from_headers(self, headers):
        """Generate a SpanContext object using the W3C Distributed Tracing headers.

        :type headers: dict
        :param headers: HTTP request headers.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from the trace context header.
        """
        if headers is None:
            return SpanContext()

        header = headers.get(_TRACEPARENT_HEADER_NAME)
        if header is None:
            return SpanContext()

        match = re.search(_TRACEPARENT_HEADER_FORMAT_RE, header)
        if not match:
            return SpanContext()

        version = match.group(1)
        trace_id = match.group(2)
        span_id = match.group(3)
        trace_options = match.group(4)

        if trace_id == '0' * 32 or span_id == '0' * 16:
            return SpanContext()

        if version == '00':
            if match.group(5):
                return SpanContext()
        if version == 'ff':
            return SpanContext()

        span_context = SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            trace_options=TraceOptions(trace_options),
            from_header=True)

        header = headers.get(_TRACESTATE_HEADER_NAME)
        if header is None:
            return span_context
        try:
            tracestate = TracestateStringFormatter().from_string(header)
            if tracestate.is_valid():
                span_context.tracestate = \
                    TracestateStringFormatter().from_string(header)
        except ValueError:
            pass
        return span_context

    def to_headers(self, span_context):
        """Convert a SpanContext object to W3C Distributed Tracing headers,
        using version 0.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :rtype: dict
        :returns: W3C Distributed Tracing headers.
        """
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        trace_options = span_context.trace_options.enabled

        # Convert the trace options
        trace_options = '01' if trace_options else '00'

        headers = {
            _TRACEPARENT_HEADER_NAME: '00-{}-{}-{}'.format(
                trace_id,
                span_id,
                trace_options
            ),
        }
        tracestate = span_context.tracestate
        if tracestate:
            headers[_TRACESTATE_HEADER_NAME] = \
                TracestateStringFormatter().to_string(tracestate)
        return headers



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import threading

from opencensus.metrics import label_key
from opencensus.metrics.export import metric_descriptor


class View(object):
    """A view defines a specific aggregation and a set of tag keys

    :type name: str
    :param name: name of the view

    :type description: str
    :param description: description of the view

    :type columns: (:class: '~opencensus.tags.tag_key.TagKey')
    :param columns: the columns that the tag keys will aggregate on for this
                    view

    :type measure: :class: '~opencensus.stats.measure.Measure'
    :param measure: the measure to be aggregated by the view

    :type aggregation: :class: '~opencensus.stats.aggregation.BaseAggregation'
    :param aggregation: the aggregation the view will support

    """

    def __init__(self, name, description, columns, measure, aggregation):
        self._name = name
        self._description = description
        self._columns = columns
        self._measure = measure
        self._aggregation = aggregation

        # Cache the converted MetricDescriptor here to avoid creating it each
        # time we convert a ViewData that realizes this View into a Metric.
        self._md_cache_lock = threading.Lock()
        self._metric_descriptor = None

    @property
    def name(self):
        """the name of the current view"""
        return self._name

    @property
    def description(self):
        """the description of the current view"""
        return self._description

    @property
    def columns(self):
        """the columns of the current view"""
        return self._columns

    @property
    def measure(self):
        """the measure of the current view"""
        return self._measure

    @property
    def aggregation(self):
        """the aggregation of the current view"""
        return self._aggregation

    def new_aggregation_data(self):
        """Get a new AggregationData for this view.

        :rtype: :class: `opencensus.status.aggregation_data.AggregationData`
        :return: A new AggregationData.
        """
        return self._aggregation.new_aggregation_data(self.measure)

    def get_metric_descriptor(self):
        """Get a MetricDescriptor for this view.

        Lazily creates a MetricDescriptor for metrics conversion.

        :rtype: :class:
                `opencensus.metrics.export.metric_descriptor.MetricDescriptor`
        :return: A converted Metric.
        """  # noqa
        with self._md_cache_lock:
            if self._metric_descriptor is None:
                self._metric_descriptor = metric_descriptor.MetricDescriptor(
                    self.name,
                    self.description,
                    self.measure.unit,
                    self.aggregation.get_metric_type(self.measure),
                    # TODO: add label key description
                    [label_key.LabelKey(tk, "") for tk in self.columns])
        return self._metric_descriptor



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from opencensus.common import utils
from opencensus.trace import attributes

_SpanData = collections.namedtuple(
    '_SpanData',
    (
        'name',
        'context',
        'span_id',
        'parent_span_id',
        'attributes',
        'start_time',
        'end_time',
        'child_span_count',
        'stack_trace',
        'annotations',
        'message_events',
        'links',
        'status',
        'same_process_as_parent_span',
        'span_kind',
    ),
)


class SpanData(_SpanData):
    """Immutable representation of all data collected by a
     :class: `~opencensus.trace.span.Span`.

    :type name: str
    :param name: The name of the span.

    :type: context: :class: `~opencensus.trace.span_context.SpanContext`
    :param context: The SpanContext of the Span

    :type span_id: int
    :param span_id: Identifier for the span, unique within a trace.

    :type parent_span_id: int
    :param parent_span_id: (Optional) Parent span id.

    :type attributes: dict
    :param attributes: Collection of attributes associated with the span.

    :type start_time: str
    :param start_time: (Optional) Start of the time interval (inclusive)
                       during which the trace data was collected from the
                       application.

    :type end_time: str
    :param end_time: (Optional) End of the time interval (inclusive) during
                     which the trace data was collected from the application.

    :type child_span_count: int
    :param child_span_count: the number of child spans that were
                            generated while the span was active.

    :type stack_trace: :class: `~opencensus.trace.stack_trace.StackTrace`
    :param stack_trace: (Optional) A call stack appearing in a trace

    :type annotations: list(:class:`opencensus.trace.time_event.Annotation`)
    :param annotations: (Optional) The list of span annotations.

    :type message_events:
        list(:class:`opencensus.trace.time_event.MessageEvent`)
    :param message_events: (Optional) The list of span message events.

    :type links: list
    :param links: (Optional) Links associated with the span. You can have up
                  to 128 links per Span.

    :type status: :class: `~opencensus.trace.status.Status`
    :param status: (Optional) An optional final status for this span.

    :type same_process_as_parent_span: bool
    :param same_process_as_parent_span: (Optional) A highly recommended but not
                                        required flag that identifies when a
                                        trace crosses a process boundary.
                                        True when the parent_span belongs to
                                        the same process as the current span.
    :type span_kind: int
    :param span_kind: (Optional) Highly recommended flag that denotes the type
                        of span (valid values defined by :class:
                        `opencensus.trace.span.SpanKind`)

    """
    __slots__ = ()


def _format_legacy_span_json(span_data):
    """
    :param SpanData span_data: SpanData object to convert
    :rtype: dict
    :return: Dictionary representing the Span
    """
    span_json = {
        'displayName': utils.get_truncatable_str(span_data.name),
        'spanId': span_data.span_id,
        'startTime': span_data.start_time,
        'endTime': span_data.end_time,
        'childSpanCount': span_data.child_span_count,
        'kind': span_data.span_kind
    }

    if span_data.parent_span_id is not None:
        span_json['parentSpanId'] = span_data.parent_span_id

    if span_data.attributes:
        span_json['attributes'] = attributes.Attributes(
            span_data.attributes).format_attributes_json()

    if span_data.stack_trace is not None:
        span_json['stackTrace'] = \
            span_data.stack_trace.format_stack_trace_json()

    formatted_time_events = []
    if span_data.annotations:
        formatted_time_events.extend(
            {'time': aa.timestamp,
             'annotation': aa.format_annotation_json()}
            for aa in span_data.annotations)
    if span_data.message_events:
        formatted_time_events.extend(
            {'time': aa.timestamp,
             'message_event': aa.format_message_event_json()}
            for aa in span_data.message_events)
    if formatted_time_events:
        span_json['timeEvents'] = {
            'timeEvent': formatted_time_events
        }

    if span_data.links:
        span_json['links'] = {
            'link': [
                link.format_link_json() for link in span_data.links]
        }

    if span_data.status is not None:
        span_json['status'] = span_data.status.format_status_json()

    if span_data.same_process_as_parent_span is not None:
        span_json['sameProcessAsParentSpan'] = \
            span_data.same_process_as_parent_span

    return span_json


def format_legacy_trace_json(span_datas):
    """Formats a list of SpanData tuples into the legacy 'trace' dictionary
    format for backwards compatibility
    :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
    :param list of opencensus.trace.span_data.SpanData span_datas:
        SpanData tuples to emit
    :rtype: dict
    :return: Legacy 'trace' dictionary representing given SpanData tuples
    """
    if not span_datas:
        return {}
    top_span = span_datas[0]
    assert isinstance(top_span, SpanData)
    trace_id = top_span.context.trace_id if top_span.context is not None \
        else None
    assert trace_id is not None
    return {
        'traceId': trace_id,
        'spans': [_format_legacy_span_json(sd) for sd in span_datas],
    }



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Summary(object):
    """Implementation of the Summary as a summary of observations.

    :type count: long
    :param count: the count of the population values.

    :type sum_data: float
    :param sum_data: the sum of the population values.

    :type snapshot: Snapshot
    :param snapshot: the values calculated over a sliding time window.
    """

    def __init__(self, count, sum_data, snapshot):
        check_count_and_sum(count, sum_data)
        self._count = count
        self._sum_data = sum_data

        if snapshot is None:
            raise ValueError('snapshot must not be none')

        self._snapshot = snapshot

    @property
    def count(self):
        """Returns the count of the population values"""
        return self._count

    @property
    def sum_data(self):
        """Returns the sum of the population values."""
        return self._sum_data

    @property
    def snapshot(self):
        """Returns the values calculated over a sliding time window."""
        return self._snapshot


class Snapshot(object):
    """Represents the summary observation of the recorded events over a
    sliding time window.

    :type count: long
    :param count: the number of values in the snapshot.

    :type sum_data: float
    :param sum_data: the sum of values in the snapshot.

    :type value_at_percentiles: ValueAtPercentile
    :param value_at_percentiles: a list of values at different percentiles
    of the distribution calculated from the current snapshot. The percentiles
    must be strictly increasing.
    """

    def __init__(self, count, sum_data, value_at_percentiles=None):
        check_count_and_sum(count, sum_data)
        self._count = count
        self._sum_data = sum_data

        if value_at_percentiles is None:
            value_at_percentiles = []

        if not isinstance(value_at_percentiles, list):
            raise ValueError('value_at_percentiles must be an '
                             'instance of list')

        self._value_at_percentiles = value_at_percentiles

    @property
    def count(self):
        """Returns the number of values in the snapshot"""
        return self._count

    @property
    def sum_data(self):
        """Returns the sum of values in the snapshot."""
        return self._sum_data

    @property
    def value_at_percentiles(self):
        """Returns a list of values at different percentiles
        of the distribution calculated from the current snapshot.
        """
        return self._value_at_percentiles


class ValueAtPercentile(object):
    """Represents the value at a given percentile of a distribution.

    :type percentile: float
    :param percentile: the percentile in the ValueAtPercentile.

    :type value: float
    :param value: the value in the ValueAtPercentile.
    """

    def __init__(self, percentile, value):

        if not 0 < percentile <= 100.0:
            raise ValueError("percentile must be in the interval (0.0, 100.0]")

        self._percentile = percentile

        if value < 0:
            raise ValueError('value must be non-negative')

        self._value = value

    @property
    def percentile(self):
        """Returns the percentile in the ValueAtPercentile"""
        return self._percentile

    @property
    def value(self):
        """Returns the value in the ValueAtPercentile"""
        return self._value


def check_count_and_sum(count, sum_data):
    if not (count is None or count >= 0):
        raise ValueError('count must be non-negative')

    if not (sum_data is None or sum_data >= 0):
        raise ValueError('sum_data must be non-negative')

    if count == 0 and sum_data != 0:
        raise ValueError('sum_data must be 0 if count is 0')



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Type(object):
    """The relationship of the current span relative to the linked span: child,
    parent, or unspecified.

    Attributes:
      TYPE_UNSPECIFIED (int): The relationship of the two spans is unknown.
      CHILD_LINKED_SPAN (int): The linked span is a child of the current span.
      PARENT_LINKED_SPAN (int): The linked span is a parent of the current
      span.
    """
    TYPE_UNSPECIFIED = 0
    CHILD_LINKED_SPAN = 1
    PARENT_LINKED_SPAN = 2


class Link(object):
    """A pointer from the current span to another span in the same trace or in
    a different trace. For example, this can be used in batching operations,
    where a single batch handler processes multiple requests from different
    traces or when the handler receives a request from a different project.

    :type trace_id: str
    :param trace_id: The [TRACE_ID] for a trace within a project.

    :type span_id: str
    :param span_id: The [SPAN_ID] for a span within a trace.

    :type type: Enum of :class:`~opencensus.trace.link.Type`
    :param type: The relationship of the current span relative to the linked
                 span.

    :type attributes: :class:`~opencensus.trace.attributes.Attributes`
    :param attributes: A set of attributes on the link. You have have up to 32
                       attributes per link.
    """
    def __init__(self, trace_id, span_id, type=None, attributes=None):
        self.trace_id = trace_id
        self.span_id = span_id

        if type is None:
            type = Type.TYPE_UNSPECIFIED

        self.type = type
        self.attributes = attributes

    def format_link_json(self):
        """Convert a Link object to json format."""
        link_json = {}
        link_json['trace_id'] = self.trace_id
        link_json['span_id'] = self.span_id
        link_json['type'] = self.type

        if self.attributes is not None:
            link_json['attributes'] = self.attributes

        return link_json



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Tracer(object):
    """Base class for Opencensus tracers.

    Subclasses of :class:`Tracer` must implement the below methods.
    """
    def finish(self):
        """End the spans and send to reporters."""
        raise NotImplementedError

    def span(self, name='span'):
        """Create a new span with the trace using the context information.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.span.Span`
        :returns: The Span object.
        """
        raise NotImplementedError

    def start_span(self, name='span'):
        """Start a span.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.span.Span`
        :returns: The Span object.
        """
        raise NotImplementedError

    def end_span(self):
        """End a span. Remove the span from the span stack, and update the
        span_id in TraceContext as the current span_id which is the peek
        element in the span stack.
        """
        raise NotImplementedError

    def current_span(self):
        """Return the current span."""
        raise NotImplementedError

    def add_attribute_to_current_span(self, attribute_key, attribute_value):
        raise NotImplementedError

    def list_collected_spans(self):
        """List collected spans."""
        raise NotImplementedError


class NullContextManager(object):
    """Empty object as a helper for faking Trace and Span when tracing is
    disabled.
    """
    def __init__(self, span_id=None, context_tracer=None):
        self.name = None
        self.span_id = span_id
        self.context_tracer = context_tracer

    def __enter__(self):
        return self  # pragma: NO COVER

    def __exit__(self, exc_type, exc_value, traceback):
        pass  # pragma: NO COVER

    def span(self, name='span'):
        return NullContextManager(context_tracer=self.context_tracer)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading

from opencensus.trace import execution_context, print_exporter
from opencensus.trace import span as trace_span
from opencensus.trace import span_data as span_data_module
from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracers import base


class ContextTracer(base.Tracer):
    """The interface for tracing a request context.

    :type span_context: :class:`~opencensus.trace.span_context.SpanContext`
    :param span_context: SpanContext encapsulates the current context within
                         the request's trace.
    """

    def __init__(self, exporter=None, span_context=None):
        if exporter is None:
            exporter = print_exporter.PrintExporter()

        if span_context is None:
            span_context = SpanContext()

        self.exporter = exporter
        self.span_context = span_context
        self.trace_id = span_context.trace_id
        self.root_span_id = span_context.span_id

        self._spans_list_condition = threading.Condition()
        # List of spans to report
        self._spans_list = []

    def finish(self):
        """Finish all spans

        :rtype: dict
        :returns: JSON format trace.
        """
        while self._spans_list:
            self.end_span()

    def span(self, name='span'):
        """Create a new span with the trace using the context information.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.span.Span`
        :returns: The Span object.
        """
        span = self.start_span(name=name)
        return span

    def start_span(self, name='span'):
        """Start a span.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.span.Span`
        :returns: The Span object.
        """
        parent_span = self.current_span()

        # If a span has remote parent span, then the parent_span.span_id
        # should be the span_id from the request header.
        if parent_span is None:
            parent_span = base.NullContextManager(
                span_id=self.span_context.span_id)

        span = trace_span.Span(
            name,
            parent_span=parent_span,
            context_tracer=self)
        with self._spans_list_condition:
            self._spans_list.append(span)
        self.span_context.span_id = span.span_id
        execution_context.set_current_span(span)
        span.start()
        return span

    def end_span(self, *args, **kwargs):
        """End a span. Update the span_id in SpanContext to the current span's
        parent span id; Update the current span.
        """
        cur_span = self.current_span()
        if cur_span is None and self._spans_list:
            cur_span = self._spans_list[-1]

        if cur_span is None:
            logging.warning('No active span, cannot do end_span.')
            return

        cur_span.finish()
        self.span_context.span_id = cur_span.parent_span.span_id if \
            cur_span.parent_span else None

        if isinstance(cur_span.parent_span, trace_span.Span):
            execution_context.set_current_span(cur_span.parent_span)
        else:
            execution_context.set_current_span(None)

        with self._spans_list_condition:
            if cur_span in self._spans_list:
                span_datas = self.get_span_datas(cur_span)
                self.exporter.export(span_datas)
                self._spans_list.remove(cur_span)

        return cur_span

    def current_span(self):
        """Return the current span."""
        current_span = execution_context.get_current_span()

        return current_span

    def list_collected_spans(self):
        return self._spans_list

    def add_attribute_to_current_span(self, attribute_key, attribute_value):
        """Add attribute to current span.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        current_span = self.current_span()
        current_span.add_attribute(attribute_key, attribute_value)

    def get_span_datas(self, span):
        """Extracts a list of SpanData tuples from a span

        :rtype: list of opencensus.trace.span_data.SpanData
        :return list of SpanData tuples
        """
        span_datas = [
            span_data_module.SpanData(
                name=ss.name,
                context=self.span_context,
                span_id=ss.span_id,
                parent_span_id=ss.parent_span.span_id if
                ss.parent_span else None,
                attributes=ss.attributes,
                start_time=ss.start_time,
                end_time=ss.end_time,
                child_span_count=len(ss.children),
                stack_trace=ss.stack_trace,
                annotations=ss.annotations,
                message_events=ss.message_events,
                links=ss.links,
                status=ss.status,
                same_process_as_parent_span=ss.same_process_as_parent_span,
                span_kind=ss.span_kind
            )
            for ss in span
        ]

        return span_datas



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import queue, range

import atexit
import logging
import threading

from opencensus.common.transports import base
from opencensus.trace import execution_context

_DEFAULT_GRACE_PERIOD = 5.0  # Seconds
_DEFAULT_MAX_BATCH_SIZE = 600
_DEFAULT_WAIT_PERIOD = 60.0  # Seconds
_WORKER_THREAD_NAME = 'opencensus.common.Worker'
_WORKER_TERMINATOR = object()

logger = logging.getLogger(__name__)


class _Worker(object):
    """A background thread that exports batches of data.

    :type exporter: :class:`~opencensus.trace.base_exporter.Exporter` or
                    :class:`~opencensus.stats.base_exporter.StatsExporter`
    :param exporter: Instance of Exporter object.

    :type grace_period: float
    :param grace_period: The amount of time to wait for pending data to
                         be submitted when the process is shutting down.

    :type max_batch_size: int
    :param max_batch_size: The maximum number of items to send at a time
                           in the background thread.

    :type wait_period: int
    :param wait_period: The amount of time to wait before sending the next
                        batch of data.
    """
    def __init__(self, exporter,
                 grace_period=_DEFAULT_GRACE_PERIOD,
                 max_batch_size=_DEFAULT_MAX_BATCH_SIZE,
                 wait_period=_DEFAULT_WAIT_PERIOD):
        self.exporter = exporter
        self._grace_period = grace_period
        self._max_batch_size = max_batch_size
        self._wait_period = wait_period
        self._queue = queue.Queue(0)
        self._lock = threading.Lock()
        self._event = threading.Event()
        self._thread = None

    @property
    def is_alive(self):
        """Returns True is the background thread is running."""
        return self._thread is not None and self._thread.is_alive()

    def _get_items(self):
        """Get multiple items from a Queue.

        Gets at least one (blocking) and at most ``max_batch_size`` items
        (non-blocking) from a given Queue. Does not mark the items as done.

        :rtype: Sequence
        :returns: A sequence of items retrieved from the queue.
        """
        items = [self._queue.get()]

        while len(items) < self._max_batch_size:
            try:
                items.append(self._queue.get_nowait())
            except queue.Empty:
                break

        return items

    def _thread_main(self):
        """The entry point for the worker thread.

        Pulls pending data off the queue and writes them in
        batches to the specified tracing backend using the exporter.
        """
        # Indicate that this thread is an exporter thread.
        # Used to suppress tracking of requests in this thread
        execution_context.set_is_exporter(True)
        quit_ = False

        while True:
            items = self._get_items()
            data = []

            for item in items:
                if item is _WORKER_TERMINATOR:
                    quit_ = True
                    # Continue processing items, don't break, try to process
                    # all items we got back before quitting.
                else:
                    data.extend(item)

            if data:
                try:
                    self.exporter.emit(data)
                except Exception:
                    logger.exception(
                        '%s failed to emit data.'
                        'Dropping %s objects from queue.',
                        self.exporter.__class__.__name__,
                        len(data))
                    pass

            for _ in range(len(items)):
                self._queue.task_done()

            # self._event is set at exit, at which point we start draining the
            # queue immediately. If self._event is unset, block for
            # self.wait_period between each batch of exports.
            self._event.wait(self._wait_period)

            if quit_:
                break

    def start(self):
        """Starts the background thread.

        Additionally, this registers a handler for process exit to attempt
        to send any pending data before shutdown.
        """
        with self._lock:
            if self.is_alive:
                return

            self._thread = threading.Thread(
                target=self._thread_main, name=_WORKER_THREAD_NAME)
            self._thread.daemon = True
            self._thread.start()
            atexit.register(self._export_pending_data)

    def stop(self):
        """Signals the background thread to stop.

        This does not terminate the background thread. It simply queues the
        stop signal. If the main process exits before the background thread
        processes the stop signal, it will be terminated without finishing
        work. The ``grace_period`` parameter will give the background
        thread some time to finish processing before this function returns.

        :rtype: bool
        :returns: True if the thread terminated. False if the thread is still
                  running.
        """
        if not self.is_alive:
            return True

        with self._lock:
            self._queue.put_nowait(_WORKER_TERMINATOR)
            self._thread.join(timeout=self._grace_period)

            success = not self.is_alive
            self._thread = None

            return success

    def _export_pending_data(self):
        """Callback that attempts to send pending data before termination."""
        if not self.is_alive:
            return
        # Stop blocking between export batches
        self._event.set()
        self.stop()

    def enqueue(self, data):
        """Queues data to be written by the background thread."""
        self._queue.put_nowait(data)

    def flush(self):
        """Submit any pending data."""
        self._queue.join()


class AsyncTransport(base.Transport):
    """Asynchronous transport that uses a background thread.

    :type exporter: :class:`~opencensus.trace.base_exporter.Exporter` or
                    :class:`~opencensus.stats.base_exporter.StatsExporter`
    :param exporter: Instance of Exporter object.

    :type grace_period: float
    :param grace_period: The amount of time to wait for pending data to
                         be submitted when the process is shutting down.

    :type max_batch_size: int
    :param max_batch_size: The maximum number of items to send at a time
                           in the background thread.

    :type wait_period: int
    :param wait_period: The amount of time to wait before sending the next
                        batch of data.
    """

    def __init__(self, exporter,
                 grace_period=_DEFAULT_GRACE_PERIOD,
                 max_batch_size=_DEFAULT_MAX_BATCH_SIZE,
                 wait_period=_DEFAULT_WAIT_PERIOD):
        self.exporter = exporter
        self.worker = _Worker(
            exporter,
            grace_period,
            max_batch_size,
            wait_period,
        )
        self.worker.start()

    def export(self, data):
        """Put the trace/stats to be exported into queue."""
        self.worker.enqueue(data)

    def flush(self):
        """Submit any pending traces/stats."""
        self.worker.flush()



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from weakref import WeakMethod
except ImportError:
    from opencensus.common.backports import WeakMethod

import calendar
import datetime
import weakref

UTF8 = 'utf-8'

# Max length is 128 bytes for a truncatable string.
MAX_LENGTH = 128

ISO_DATETIME_REGEX = '%Y-%m-%dT%H:%M:%S.%fZ'


def get_truncatable_str(str_to_convert):
    """Truncate a string if exceed limit and record the truncated bytes
    count.
    """
    truncated, truncated_byte_count = check_str_length(
        str_to_convert, MAX_LENGTH)

    result = {
        'value': truncated,
        'truncated_byte_count': truncated_byte_count,
    }
    return result


def check_str_length(str_to_check, limit=MAX_LENGTH):
    """Check the length of a string. If exceeds limit, then truncate it.

    :type str_to_check: str
    :param str_to_check: String to check.

    :type limit: int
    :param limit: The upper limit of the length.

    :rtype: tuple
    :returns: The string it self if not exceeded length, or truncated string
              if exceeded and the truncated byte count.
    """
    str_bytes = str_to_check.encode(UTF8)
    str_len = len(str_bytes)
    truncated_byte_count = 0

    if str_len > limit:
        truncated_byte_count = str_len - limit
        str_bytes = str_bytes[:limit]

    result = str(str_bytes.decode(UTF8, errors='ignore'))

    return (result, truncated_byte_count)


def to_iso_str(ts=None):
    """Get an ISO 8601 string for a UTC datetime."""
    if ts is None:
        ts = datetime.datetime.utcnow()
    return ts.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def timestamp_to_microseconds(timestamp):
    """Convert a timestamp string into a microseconds value
    :param timestamp
    :return time in microseconds
    """
    timestamp_str = datetime.datetime.strptime(timestamp, ISO_DATETIME_REGEX)
    epoch_time_secs = calendar.timegm(timestamp_str.timetuple())
    epoch_time_mus = epoch_time_secs * 1e6 + timestamp_str.microsecond
    return epoch_time_mus


def iuniq(ible):
    """Get an iterator over unique items of `ible`."""
    items = set()
    for item in ible:
        if item not in items:
            items.add(item)
            yield item


def uniq(ible):
    """Get a list of unique items of `ible`."""
    return list(iuniq(ible))


def window(ible, length):
    """Split `ible` into multiple lists of length `length`.

    >>> list(window(range(5), 2))
    [[0, 1], [2, 3], [4]]
    """
    if length <= 0:  # pragma: NO COVER
        raise ValueError
    ible = iter(ible)
    while True:
        elts = [xx for ii, xx in zip(range(length), ible)]
        if elts:
            yield elts
        else:
            break


def get_weakref(func):
    """Get a weak reference to bound or unbound `func`.

    If `func` is unbound (i.e. has no __self__ attr) get a weakref.ref,
    otherwise get a wrapper that simulates weakref.ref.
    """
    if func is None:
        raise ValueError
    if not hasattr(func, '__self__'):
        return weakref.ref(func)
    return WeakMethod(func)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from opencensus.metrics.export.metric_producer import MetricProducer
from opencensus.stats.stats_recorder import StatsRecorder
from opencensus.stats.view_manager import ViewManager


class _Stats(MetricProducer):
    """Stats defines a View Manager and a Stats Recorder in order for the
    collection of Stats
    """

    def __init__(self):
        self.stats_recorder = StatsRecorder()
        self.view_manager = ViewManager()

    def get_metrics(self):
        """Get a Metric for each of the view manager's registered views.

        Convert each registered view's associated `ViewData` into a `Metric` to
        be exported, using the current time for metric conversions.

        :rtype: Iterator[:class: `opencensus.metrics.export.metric.Metric`]
        """
        return self.view_manager.measure_to_view_map.get_metrics(
            datetime.utcnow())


stats = _Stats()



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging

from opencensus.common import utils
from opencensus.common.schedule import PeriodicTask
from opencensus.trace import execution_context

logger = logging.getLogger(__name__)

DEFAULT_INTERVAL = 60
GRACE_PERIOD = 5


class TransportError(Exception):
    pass


class PeriodicMetricTask(PeriodicTask):
    """Thread that periodically calls a given function.

    :type interval: int or float
    :param interval: Seconds between calls to the function.

    :type function: function
    :param function: The function to call.

    :type args: list
    :param args: The args passed in while calling `function`.

    :type kwargs: dict
    :param args: The kwargs passed in while calling `function`.

    :type name: str
    :param name: The source of the worker. Used for naming.
    """

    daemon = True

    def __init__(
        self,
        interval=None,
        function=None,
        args=None,
        kwargs=None,
        name=None
    ):
        if interval is None:
            interval = DEFAULT_INTERVAL

        self.func = function
        self.args = args
        self.kwargs = kwargs

        def func(*aa, **kw):
            try:
                return self.func(*aa, **kw)
            except TransportError as ex:
                logger.exception(ex)
                self.cancel()
            except Exception as ex:
                logger.exception("Error handling metric export: {}".format(ex))

        super(PeriodicMetricTask, self).__init__(
            interval, func, args, kwargs, '{} Worker'.format(name)
        )

    def run(self):
        # Indicate that this thread is an exporter thread.
        # Used to suppress tracking of requests in this thread
        execution_context.set_is_exporter(True)
        super(PeriodicMetricTask, self).run()

    def close(self):
        try:
            # Suppress request tracking on flush
            execution_context.set_is_exporter(True)
            self.func(*self.args, **self.kwargs)
            execution_context.set_is_exporter(False)
        except Exception as ex:
            logger.exception("Error handling metric flush: {}".format(ex))
        self.cancel()


def get_exporter_thread(metric_producers, exporter, interval=None):
    """Get a running task that periodically exports metrics.

    Get a `PeriodicTask` that periodically calls:

        export(itertools.chain(*all_gets))

    where all_gets is the concatenation of all metrics produced by the metric
    producers in metric_producers, each calling metric_producer.get_metrics()

    :type metric_producers:
    list(:class:`opencensus.metrics.export.metric_producer.MetricProducer`)
    :param metric_producers: The list of metric producers to use to get metrics

    :type exporter: :class:`opencensus.stats.base_exporter.MetricsExporter`
    :param exporter: The exporter to use to export metrics.

    :type interval: int or float
    :param interval: Seconds between export calls.

    :rtype: :class:`PeriodicTask`
    :return: A running thread responsible calling the exporter.

    """
    weak_gets = [utils.get_weakref(producer.get_metrics)
                 for producer in metric_producers]
    weak_export = utils.get_weakref(exporter.export_metrics)

    def export_all():
        all_gets = []
        for weak_get in weak_gets:
            get = weak_get()
            if get is None:
                raise TransportError("Metric producer is not available")
            all_gets.append(get())
        export = weak_export()
        if export is None:
            raise TransportError("Metric exporter is not available")

        export(itertools.chain(*all_gets))

    tt = PeriodicMetricTask(
        interval,
        export_all,
        name=exporter.__class__.__name__
    )
    tt.start()
    return tt



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common.runtime_context import RuntimeContext
from opencensus.trace.tracers import noop_tracer

_attrs_slot = RuntimeContext.register_slot('attrs', lambda: {})
_current_span_slot = RuntimeContext.register_slot('current_span', None)
_exporter_slot = RuntimeContext.register_slot('is_exporter', False)
_tracer_slot = RuntimeContext.register_slot('tracer', noop_tracer.NoopTracer())


def is_exporter():
    return RuntimeContext.is_exporter


def set_is_exporter(is_exporter):
    RuntimeContext.is_exporter = is_exporter


def get_opencensus_tracer():
    """Get the opencensus tracer from runtime context."""
    return RuntimeContext.tracer


def set_opencensus_tracer(tracer):
    """Add the tracer to runtime context."""
    RuntimeContext.tracer = tracer


def set_opencensus_attr(attr_key, attr_value):
    attrs = RuntimeContext.attrs.copy()
    attrs[attr_key] = attr_value
    RuntimeContext.attrs = attrs


def set_opencensus_attrs(attrs):
    RuntimeContext.attrs = attrs


def get_opencensus_attr(attr_key):
    return RuntimeContext.attrs.get(attr_key)


def get_opencensus_attrs():
    return RuntimeContext.attrs


def get_current_span():
    return RuntimeContext.current_span


def set_current_span(current_span):
    RuntimeContext.current_span = current_span


def get_opencensus_full_context():
    attrs = RuntimeContext.attrs
    current_span = RuntimeContext.current_span
    tracer = RuntimeContext.tracer
    return tracer, current_span, attrs


def set_opencensus_full_context(tracer, span, attrs):
    set_opencensus_tracer(tracer)
    set_current_span(span)
    set_opencensus_attrs(attrs or {})


def clean():
    _attrs_slot.clear()
    _current_span_slot.clear()
    _tracer_slot.clear()


def clear():
    """Clear the context, used in test."""
    clean()



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import OrderedDict

_KEY_WITHOUT_VENDOR_FORMAT = r'[a-z][_0-9a-z\-\*\/]{0,255}'
_KEY_WITH_VENDOR_FORMAT = \
    r'[a-z][_0-9a-z\-\*\/]{0,240}@[a-z][_0-9a-z\-\*\/]{0,13}'
_KEY_FORMAT = _KEY_WITHOUT_VENDOR_FORMAT + '|' + _KEY_WITH_VENDOR_FORMAT
_VALUE_FORMAT = \
    r'[\x20-\x2b\x2d-\x3c\x3e-\x7e]{0,255}[\x21-\x2b\x2d-\x3c\x3e-\x7e]'

_KEY_VALIDATION_RE = re.compile('^' + _KEY_FORMAT + '$')
_VALUE_VALIDATION_RE = re.compile('^' + _VALUE_FORMAT + '$')


class Tracestate(OrderedDict):
    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise ValueError('key must be an instance of str')
        if not re.match(_KEY_VALIDATION_RE, key):
            raise ValueError('illegal key provided')
        if not isinstance(value, str):
            raise ValueError('value must be an instance of str')
        if not re.match(_VALUE_VALIDATION_RE, value):
            raise ValueError('illegal value provided')
        super(Tracestate, self).__setitem__(key, value)

    def append(self, key, value):
        if self.get(key):
            del self[key]
        self[key] = value

    # make this an optional choice instead of enforcement during put/update
    # if the tracestate value size is bigger than 512 characters, the tracer
    # CAN decide to forward the tracestate
    def is_valid(self):
        if len(self) == 0:
            return False
        # there can be a maximum of 32 list-members in a list
        if len(self) > 32:
            return False
        return True

    def prepend(self, key, value):
        self[key] = value
        if hasattr(self, 'move_to_end'):
            self.move_to_end(key, last=False)
        else:  # less performant way for Python 2.x
            copy = OrderedDict(self)
            self.clear()
            self[key] = value
            self.update(copy)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.metrics.export import metric_descriptor


class Metric(object):
    """A collection of time series data and label metadata.

    This class implements the spec for v1 Metrics as of opencensus-proto
    release v0.1.0. See opencensus-proto for details:

    https://github.com/census-instrumentation/opencensus-proto/blob/v0.1.0/src/opencensus/proto/metrics/v1/metrics.proto#L35

    Defines a Metric which has one or more timeseries.

    :type descriptor: class: '~opencensus.metrics.export.metric_descriptor.MetricDescriptor'
    :param descriptor: The metric's descriptor.

    :type timeseries: list(:class: '~opencensus.metrics.export.time_series.TimeSeries')
    :param timeseries: One or more timeseries for a single metric, where each
    timeseries has one or more points.
    """  # noqa

    def __init__(self, descriptor, time_series):
        if not time_series:
            raise ValueError("time_series must not be empty or null")
        if descriptor is None:
            raise ValueError("descriptor must not be null")
        self._time_series = time_series
        self._descriptor = descriptor
        self._check_type()

    def __repr__(self):
        return ('{}(time_series={}, descriptor.name="{}")'
                .format(
                    type(self).__name__,
                    "<{} TimeSeries>".format(len(self.time_series)),
                    self.descriptor.name,
                ))

    @property
    def time_series(self):
        return self._time_series

    @property
    def descriptor(self):
        return self._descriptor

    def _check_type(self):
        """Check that point value types match the descriptor type."""
        check_type = metric_descriptor.MetricDescriptorType.to_type_class(
            self.descriptor.type)
        for ts in self.time_series:
            if not ts.check_points_type(check_type):
                raise ValueError("Invalid point value type")

    def _check_start_timestamp(self):
        """Check that starting timestamp exists for cumulative metrics."""
        if self.descriptor.type in (
                metric_descriptor.MetricDescriptorType.CUMULATIVE_INT64,
                metric_descriptor.MetricDescriptorType.CUMULATIVE_DOUBLE,
                metric_descriptor.MetricDescriptorType.CUMULATIVE_DISTRIBUTION,
        ):
            for ts in self.time_series:
                if ts.start_timestamp is None:
                    raise ValueError("time_series.start_timestamp must exist "
                                     "for cumulative metrics")



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def is_legal_chars(value):
    return all(32 <= ord(char) <= 126 for char in value)


def is_valid_tag_name(name):
    """Checks if the name of a tag key is valid

    :type name: str
    :param name: name to check

    :rtype: bool
    :returns: True if it valid, else returns False
    """
    return is_legal_chars(name) if 0 < len(name) <= 255 else False


def is_valid_tag_value(value):
    """Checks if the value is valid

    :type value: str
    :param value: the value to be checked

    :rtype: bool
    :returns: True if valid, if not, False.

    """
    return is_legal_chars(value) if len(value) <= 255 else False



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

import weakref


class WeakMethod(weakref.ref):  # pragma: NO COVER
    """
    A custom `weakref.ref` subclass which simulates a weak reference to
    a bound method, working around the lifetime problem of bound methods.

    This is a copy of the WeakMethod class that ships with weakref in the
    python 3.7 standard library, adapted to work in 2.6. See:
    https://github.com/python/cpython/blob/a31f4cc881992e84d351957bd9ac1a92f882fa39/Lib/weakref.py#L36-L87
    """  # noqa

    __slots__ = "_func_ref", "_meth_type", "_alive", "__weakref__"

    def __new__(cls, meth, callback=None):
        try:
            obj = meth.__self__
            func = meth.__func__
        except AttributeError:
            error = TypeError("argument should be a bound method, not {}"
                              .format(type(meth)))
            six.raise_from(error, None)

        def _cb(arg):
            # The self-weakref trick is needed to avoid creating a reference
            # cycle.
            self = self_wr()
            if self._alive:
                self._alive = False
                if callback is not None:
                    callback(self)
        self = weakref.ref.__new__(cls, obj, _cb)
        self._func_ref = weakref.ref(func, _cb)
        self._meth_type = type(meth)
        self._alive = True
        self_wr = weakref.ref(self)
        return self

    def __call__(self):
        obj = super(WeakMethod, self).__call__()
        func = self._func_ref()
        if obj is None or func is None:
            return None
        return self._meth_type(func, obj)

    def __eq__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is other
            return (weakref.ref.__eq__(self, other)
                    and self._func_ref == other._func_ref)
        return False

    def __ne__(self, other):
        if isinstance(other, WeakMethod):
            if not self._alive or not other._alive:
                return self is not other
            return (weakref.ref.__ne__(self, other)
                    or self._func_ref != other._func_ref)
        return True

    __hash__ = weakref.ref.__hash__



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common import utils
from opencensus.stats import execution_context
from opencensus.stats.measure_to_view_map import MeasureToViewMap


class ViewManager(object):
    """View Manager allows the registering of Views for collecting stats
    and receiving stats data as View Data"""
    def __init__(self):
        self.time = utils.to_iso_str()
        if execution_context.get_measure_to_view_map() == {}:
            execution_context.set_measure_to_view_map(MeasureToViewMap())

        self._measure_view_map = execution_context.get_measure_to_view_map()

    @property
    def measure_to_view_map(self):
        """the current measure to view map for the View Manager"""
        return self._measure_view_map

    def register_view(self, view):
        """registers the given view"""
        self.measure_to_view_map.register_view(view=view, timestamp=self.time)

    def get_view(self, view_name):
        """gets the view given the view name """
        return self.measure_to_view_map.get_view(view_name=view_name,
                                                 timestamp=self.time)

    def get_all_exported_views(self):
        """returns all of the exported views for the current measure to view
        map"""
        return self.measure_to_view_map.exported_views

    def register_exporter(self, exporter):
        """register the exporter"""
        self.measure_to_view_map.exporters.append(exporter)

    def unregister_exporter(self, exporter):
        """unregister the exporter"""
        self.measure_to_view_map.exporters.remove(exporter)



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import re
from copy import copy

logger = logging.getLogger(__name__)


OC_RESOURCE_TYPE = 'OC_RESOURCE_TYPE'
OC_RESOURCE_LABELS = 'OC_RESOURCE_LABELS'

# Matches anything outside ASCII 32-126 inclusive
_NON_PRINTABLE_ASCII = re.compile(
    r'[^ !"#$%&\'()*+,\-./:;<=>?@\[\\\]^_`{|}~0-9a-zA-Z]')

# Label key/value tokens, may be quoted
_WORD_RES = r'(\'[^\']*\'|"[^"]*"|[^\s,=]+)'

_KV_RE = re.compile(r"""
    \s*                 # ignore leading spaces
    (?P<key>{word_re})  # capture the key word
    \s*=\s*
    (?P<val>{word_re})  # capture the value word
    \s*                 # ignore trailing spaces
    """.format(word_re=_WORD_RES), re.VERBOSE)

_LABELS_RE = re.compile(r"""
    ^\s*{word_re}\s*=\s*{word_re}\s*     # _KV_RE without the named groups
    (,\s*{word_re}\s*=\s*{word_re}\s*)*  # more KV pairs, comma delimited
    $
    """.format(word_re=_WORD_RES), re.VERBOSE)

_UNQUOTE_RE = re.compile(r'^([\'"]?)([^\1]*)(\1)$')


def merge_resources(resource_list):
    """Merge multiple resources to get a new resource.

    Resources earlier in the list take precedence: if multiple resources share
    a label key, use the value from the first resource in the list with that
    key. The combined resource's type will be the first non-null type in the
    list.

    :type resource_list: list(:class:`Resource`)
    :param resource_list: The list of resources to combine.

    :rtype: :class:`Resource`
    :return: The new combined resource.
    """
    if not resource_list:
        raise ValueError
    rtype = None
    for rr in resource_list:
        if rr.type:
            rtype = rr.type
            break
    labels = {}
    for rr in reversed(resource_list):
        labels.update(rr.labels)
    return Resource(rtype, labels)


def check_ascii_256(string):
    """Check that `string` is printable ASCII and at most 256 chars.

    Raise a `ValueError` if this check fails. Note that `string` itself doesn't
    have to be ASCII-encoded.

    :type string: str
    :param string: The string to check.
    """
    if string is None:
        return
    if len(string) > 256:
        raise ValueError("Value is longer than 256 characters")
    bad_char = _NON_PRINTABLE_ASCII.search(string)
    if bad_char:
        raise ValueError(u'Character "{}" at position {} is not printable '
                         'ASCII'
                         .format(
                             string[bad_char.start():bad_char.end()],
                             bad_char.start()))


class Resource(object):
    """A description of the entity for which signals are reported.

    `type_` and `labels`' keys and values should contain only printable ASCII
    and should be at most 256 characters.

    See:
        https://github.com/census-instrumentation/opencensus-specs/blob/master/resource/Resource.md

    :type type_: str
    :param type_: The resource type identifier.

    :type labels: dict
    :param labels: Key-value pairs that describe the entity.
    """  # noqa

    def __init__(self, type_=None, labels=None):
        if type_ is not None and not type_:
            raise ValueError("Resource type must not be empty")
        check_ascii_256(type_)
        if labels is None:
            labels = {}
        for key, value in labels.items():
            if not key:
                raise ValueError("Resource key must not be null or empty")
            if value is None:
                raise ValueError("Resource value must not be null")
            check_ascii_256(key)
            check_ascii_256(value)

        self.type = type_
        self.labels = copy(labels)

    def get_type(self):
        """Get this resource's type.

        :rtype: str
        :return: The resource's type.
        """
        return self.type

    def get_labels(self):
        """Get this resource's labels.

        :rtype: dict
        :return: The resource's label dict.
        """
        return copy(self.labels)

    def merge(self, other):
        """Get a copy of this resource combined with another resource.

        The combined resource will have the union of both resources' labels,
        keeping this resource's label values if they conflict.

        :type other: :class:`Resource`
        :param other: The other resource to merge.

        :rtype: :class:`Resource`
        :return: The new combined resource.
        """
        return merge_resources([self, other])


def unquote(string):
    """Strip quotes surrounding `string` if they exist.

    >>> unquote('abc')
    'abc'
    >>> unquote('"abc"')
    'abc'
    >>> unquote("'abc'")
    'abc'
    >>> unquote('"a\\'b\\'c"')
    "a'b'c"
    """
    return _UNQUOTE_RE.sub(r'\2', string)


def parse_labels(labels_str):
    """Parse label keys and values following the Resource spec.

    >>> parse_labels("k=v")
    {'k': 'v'}
    >>> parse_labels("k1=v1, k2=v2")
    {'k1': 'v1', 'k2': 'v2'}
    >>> parse_labels("k1='v1,=z1'")
    {'k1': 'v1,=z1'}
    """
    if not _LABELS_RE.match(labels_str):
        return None
    labels = {}
    for kv in _KV_RE.finditer(labels_str):
        gd = kv.groupdict()
        key = unquote(gd['key'])
        if key in labels:
            logger.warning('Duplicate label key "%s"', key)
        labels[key] = unquote(gd['val'])
    return labels


def get_from_env():
    """Get a Resource from environment variables.

    :rtype: :class:`Resource`
    :return: A resource with type and labels from the environment.
    """
    type_env = os.getenv(OC_RESOURCE_TYPE)
    if type_env is None:
        return None
    type_env = type_env.strip()

    labels_env = os.getenv(OC_RESOURCE_LABELS)
    if labels_env is None:
        return Resource(type_env)

    labels = parse_labels(labels_env)

    return Resource(type_env, labels)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.trace import blank_span as trace_span
from opencensus.trace import trace_options
from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracers import base


class NoopTracer(base.Tracer):
    """No-op implementation of the :class:`Tracer` interface, all methods are
    no-ops. Should be used when tracing is not enabled or not sampled.
    """

    def __init__(self):

        self.span_context = SpanContext(
            trace_options=trace_options.TraceOptions(0)
        )

    def finish(self):
        """End spans and send to reporter."""
        return None

    def span(self, name='span'):
        """Create a new span with the trace using the context information.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.trace_span.Span`
        :returns: The Span object.
        """

        span = self.start_span(name=name)
        return span

    def start_span(self, name='span'):
        """Start a span.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.trace_span.Span`
        :returns: The Span object.
        """
        span = trace_span.BlankSpan(name, context_tracer=self)
        return span

    def end_span(self):
        """End a span. Remove the span from the span stack, and update the
        span_id in TraceContext as the current span_id which is the peek
        element in the span stack.
        """
        pass

    def current_span(self):
        """Return the current span."""
        return trace_span.BlankSpan()

    def add_attribute_to_current_span(self, attribute_key, attribute_value):
        """Add attribute to current span.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        return

    def list_collected_spans(self):
        """List collected spans."""
        return None



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging

log = logging.getLogger(__name__)


def trace_integrations(integrations, tracer=None):
    """Enable tracing on the selected integrations.
    :type integrations: list
    :param integrations: The integrations to be traced.
    """
    integrated = []

    for item in integrations:
        module_name = 'opencensus.ext.{}.trace'.format(item)
        try:
            module = importlib.import_module(module_name)
            module.trace_integration(tracer=tracer)
            integrated.append(item)
        except Exception as e:
            log.warning('Failed to integrate module: {}'.format(module_name))
            log.warning('{}'.format(e))

    return integrated



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BucketBoundaries(object):
    """The bucket boundaries for a histogram

    :type boundaries: list(float)
    :param boundaries: boundaries for the buckets in the underlying histogram

    """
    def __init__(self, boundaries=None):
        self._boundaries = list(boundaries or [])

    @property
    def boundaries(self):
        """the current boundaries"""
        return self._boundaries

    def is_valid_boundaries(self, boundaries):
        """checks if the boundaries are in ascending order"""
        if boundaries is not None:
            min_ = boundaries[0]
            for value in boundaries:
                if value < min_:
                    return False
                else:
                    min_ = value
            return True
        return False



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.tags.validation import is_valid_tag_name

_TAG_NAME_ERROR = \
    'tag name must not be empty,' \
    'no longer than 255 characters and of ascii values between 32 - 126'


class TagKey(str):
    """A tag key with a property name"""

    def __new__(cls, name):
        """Create and return a new tag key

        :type name: str
        :param name: The name of the key
        :return: TagKey
        """
        if not isinstance(name, cls):
            if not is_valid_tag_name(name):
                raise ValueError(_TAG_NAME_ERROR)
        return super(TagKey, cls).__new__(cls, name)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from opencensus.trace.span_context import SpanContext
from opencensus.trace.trace_options import TraceOptions

_TRACE_CONTEXT_HEADER_NAME = 'X-Cloud-Trace-Context'
_TRACE_CONTEXT_HEADER_FORMAT = r'([0-9a-f]{32})(\/([\d]{0,20}))?(;o=(\d+))?'
_TRACE_CONTEXT_HEADER_RE = re.compile(_TRACE_CONTEXT_HEADER_FORMAT)
_TRACE_ID_DELIMETER = '/'
_SPAN_ID_DELIMETER = ';'


class GoogleCloudFormatPropagator(object):
    """This class is for converting the trace header in google cloud format
    and generate a SpanContext, or converting a SpanContext to a google cloud
    format header. Later we will add implementation for supporting other
    format like binary format and zipkin, opencensus format.
    """
    def from_header(self, header):
        """Generate a SpanContext object using the trace context header.
        The value of enabled parsed from header is int. Need to convert to
        bool.

        :type header: str
        :param header: Trace context header which was extracted from the HTTP
                       request headers.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from the trace context header.
        """
        if header is None:
            return SpanContext()

        try:
            match = re.search(_TRACE_CONTEXT_HEADER_RE, header)
        except TypeError:
            logging.warning(
                'Header should be str, got %s. Cannot parse the header.',
                header.__class__.__name__)
            raise

        if match:
            trace_id = match.group(1)
            span_id = match.group(3)
            trace_options = match.group(5)

            if trace_options is None:
                trace_options = 1

            if span_id:
                span_id = '{:016x}'.format(int(span_id))

            span_context = SpanContext(
                trace_id=trace_id,
                span_id=span_id,
                trace_options=TraceOptions(trace_options),
                from_header=True)
            return span_context
        else:
            logging.warning(
                'Cannot parse the header %s, generate a new context instead.',
                header)
            return SpanContext()

    def from_headers(self, headers):
        """Generate a SpanContext object using the trace context header.

        :type headers: dict
        :param headers: HTTP request headers.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from the trace context header.
        """
        if headers is None:
            return SpanContext()
        header = headers.get(_TRACE_CONTEXT_HEADER_NAME)
        if header is None:
            return SpanContext()
        return self.from_header(header)

    def to_header(self, span_context):
        """Convert a SpanContext object to header string.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :rtype: str
        :returns: A trace context header string in google cloud format.
        """
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        trace_options = span_context.trace_options.trace_options_byte

        header = '{}/{};o={}'.format(
            trace_id,
            int(span_id, 16),
            int(trace_options))
        return header

    def to_headers(self, span_context):
        """Convert a SpanContext object to HTTP request headers.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :rtype: dict
        :returns: Trace context headers in google cloud format.
        """
        return {
            _TRACE_CONTEXT_HEADER_NAME: self.to_header(span_context),
        }



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export the trace spans to a local file."""

import json

from opencensus.common.transports import sync
from opencensus.trace import base_exporter, span_data

DEFAULT_FILENAME = 'opencensus-traces.json'


class FileExporter(base_exporter.Exporter):
    """
    :type file_name: str
    :param file_name: The name of the output file.

    :type transport: :class:`type`
    :param transport: Class for creating new transport objects. It should
                      extend from the base_exporter :class:`.Transport` type
                      and implement :meth:`.Transport.export`. Defaults to
                      :class:`.SyncTransport`. The other option is
                      :class:`.AsyncTransport`.

    :type file_mode: str
    :param file_mode: The file mode to open the output file with.
                      Defaults to w+

    """

    def __init__(self, file_name=DEFAULT_FILENAME,
                 transport=sync.SyncTransport,
                 file_mode='w+'):
        self.file_name = file_name
        self.transport = transport(self)
        self.file_mode = file_mode

    def emit(self, span_datas):
        """
        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to emit
        """
        with open(self.file_name, self.file_mode) as file:
            # convert to the legacy trace json for easier refactoring
            # TODO: refactor this to use the span data directly
            legacy_trace_json = span_data.format_legacy_trace_json(span_datas)
            trace_str = json.dumps(legacy_trace_json)
            file.write(trace_str)

    def export(self, span_datas):
        """
        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to export
        """
        self.transport.export(span_datas)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from opencensus.common import utils
from opencensus.tags import TagContext

logger = logging.getLogger(__name__)


class MeasurementMap(object):
    """Measurement Map is a map from Measures to measured values
    to be recorded at the same time

    :type measure_to_view_map: :class: '~opencensus.stats.measure_to_view_map.
                                        MeasureToViewMap'
    :param measure_to_view_map: the measure to view map that will store the
                                recorded stats with tags

    :type: attachments: dict
    :param attachments: the contextual information about the attachment value.

    """
    def __init__(self, measure_to_view_map, attachments=None):
        self._measurement_map = {}
        self._measure_to_view_map = measure_to_view_map
        self._attachments = attachments
        # If the user tries to record a negative value for any measurement,
        # refuse to record all measurements from this map. Recording negative
        # measurements will become an error in a later release.
        self._invalid = False

    @property
    def measurement_map(self):
        """the current measurement map"""
        return self._measurement_map

    @property
    def measure_to_view_map(self):
        """the current measure to view map for the measurement map"""
        return self._measure_to_view_map

    @property
    def attachments(self):
        """the current contextual information about the attachment value."""
        return self._attachments

    def measure_int_put(self, measure, value):
        """associates the measure of type Int with the given value"""
        if value < 0:
            # Should be an error in a later release.
            logger.warning("Cannot record negative values")
        self._measurement_map[measure] = value

    def measure_float_put(self, measure, value):
        """associates the measure of type Float with the given value"""
        if value < 0:
            # Should be an error in a later release.
            logger.warning("Cannot record negative values")
        self._measurement_map[measure] = value

    def measure_put_attachment(self, key, value):
        """Associate the contextual information of an Exemplar to this MeasureMap
            Contextual information is represented as key - value string pairs.
            If this method is called multiple times with the same key,
            only the last value will be kept.
        """
        if self._attachments is None:
            self._attachments = dict()

        if key is None or not isinstance(key, str):
            raise TypeError('attachment key should not be '
                            'empty and should be a string')
        if value is None or not isinstance(value, str):
            raise TypeError('attachment value should not be '
                            'empty and should be a string')

        self._attachments[key] = value

    def record(self, tags=None):
        """records all the measures at the same time with a tag_map.
        tag_map could either be explicitly passed to the method, or implicitly
        read from current runtime context.
        """
        if tags is None:
            tags = TagContext.get()
        if self._invalid:
            logger.warning("Measurement map has included negative value "
                           "measurements, refusing to record")
            return
        for measure, value in self.measurement_map.items():
            if value < 0:
                self._invalid = True
                logger.warning("Dropping values, value to record must be "
                               "non-negative")
                logger.info("Measure '{}' has negative value ({}), refusing "
                            "to record measurements from {}"
                            .format(measure.name, value, self))
                return

        self.measure_to_view_map.record(
                tags=tags,
                measurement_map=self.measurement_map,
                timestamp=utils.to_iso_str(),
                attachments=self.attachments
        )



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging

from opencensus.metrics.export import point, value
from opencensus.stats import bucket_boundaries

logger = logging.getLogger(__name__)


class SumAggregationData(object):
    """Sum Aggregation Data is the aggregated data for the Sum aggregation

    :type value_type: class that is either
        :class:`opencensus.metrics.export.value.ValueDouble` or
        :class:`opencensus.metrics.export.value.ValueLong`
    :param value_type: the type of value to be used when creating a point
    :type sum_data: int or float
    :param sum_data: represents the initial aggregated sum

    """

    def __init__(self, value_type, sum_data):
        self._value_type = value_type
        self._sum_data = sum_data

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.sum_data,
                ))

    def add_sample(self, value, timestamp=None, attachments=None):
        """Allows the user to add a sample to the Sum Aggregation Data
        The value of the sample is then added to the current sum data
        """
        self._sum_data += value

    @property
    def sum_data(self):
        """The current sum data"""
        return self._sum_data

    @property
    def value_type(self):
        """The value type to use when creating the point"""
        return self._value_type

    def to_point(self, timestamp):
        """Get a Point conversion of this aggregation.

        :type timestamp: :class: `datetime.datetime`
        :param timestamp: The time to report the point as having been recorded.

        :rtype: :class: `opencensus.metrics.export.point.Point`
        :return: a Point with value equal to `sum_data` and of type
            `_value_type`.
        """
        return point.Point(self._value_type(self.sum_data), timestamp)


class CountAggregationData(object):
    """Count Aggregation Data is the count value of aggregated data

    :type count_data: long
    :param count_data: represents the initial aggregated count

    """

    def __init__(self, count_data):
        self._count_data = count_data

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.count_data,
                ))

    def add_sample(self, value, timestamp=None, attachments=None):
        """Adds a sample to the current Count Aggregation Data and adds 1 to
        the count data"""
        self._count_data = self._count_data + 1

    @property
    def count_data(self):
        """The current count data"""
        return self._count_data

    def to_point(self, timestamp):
        """Get a Point conversion of this aggregation.

        :type timestamp: :class: `datetime.datetime`
        :param timestamp: The time to report the point as having been recorded.

        :rtype: :class: `opencensus.metrics.export.point.Point`
        :return: a :class: `opencensus.metrics.export.value.ValueLong`-valued
        Point with value equal to `count_data`.
        """
        return point.Point(value.ValueLong(self.count_data), timestamp)


class DistributionAggregationData(object):
    """Distribution Aggregation Data refers to the distribution stats of
    aggregated data

    :type mean_data: float
    :param mean_data: the mean value of the distribution

    :type count_data: int
    :param count_data: the count value of the distribution

    :type sum_of_sqd_deviations: float
    :param sum_of_sqd_deviations: the sum of the sqd deviations from the mean

    :type counts_per_bucket: list(int)
    :param counts_per_bucket: the number of occurrences per bucket

    :type exemplars: list(Exemplar)
    :param: exemplars: the exemplars associated with histogram buckets.

    :type bounds: list(float)
    :param bounds: the histogram distribution of the values

    """

    def __init__(self,
                 mean_data,
                 count_data,
                 sum_of_sqd_deviations,
                 counts_per_bucket=None,
                 bounds=None,
                 exemplars=None):
        if bounds is None and exemplars is not None:
            raise ValueError
        if exemplars is not None and len(exemplars) != len(bounds) + 1:
            raise ValueError

        self._mean_data = mean_data
        self._count_data = count_data
        self._sum_of_sqd_deviations = sum_of_sqd_deviations

        if bounds is None:
            bounds = []
            self._exemplars = None
        else:
            assert bounds == list(sorted(set(bounds)))
            assert all(bb > 0 for bb in bounds)
            if exemplars is None:
                self._exemplars = {ii: None for ii in range(len(bounds) + 1)}
            else:
                self._exemplars = {ii: ex for ii, ex in enumerate(exemplars)}
        self._bounds = (bucket_boundaries.BucketBoundaries(boundaries=bounds)
                        .boundaries)

        if counts_per_bucket is None:
            counts_per_bucket = [0 for ii in range(len(bounds) + 1)]
        else:
            assert all(cc >= 0 for cc in counts_per_bucket)
            assert len(counts_per_bucket) == len(bounds) + 1
        self._counts_per_bucket = counts_per_bucket

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.count_data,
                ))

    @property
    def mean_data(self):
        """The current mean data"""
        return self._mean_data

    @property
    def count_data(self):
        """The current count data"""
        return self._count_data

    @property
    def sum_of_sqd_deviations(self):
        """The current sum of squared deviations from the mean"""
        return self._sum_of_sqd_deviations

    @property
    def counts_per_bucket(self):
        """The current counts per bucket for the distribution"""
        return self._counts_per_bucket

    @property
    def exemplars(self):
        """The current counts per bucket for the distribution"""
        return self._exemplars

    @property
    def bounds(self):
        """The current bounds for the distribution"""
        return self._bounds

    @property
    def sum(self):
        """The sum of the current distribution"""
        return self._mean_data * self._count_data

    @property
    def variance(self):
        """The variance of the current distribution"""
        if self._count_data <= 1:
            return 0
        return self.sum_of_sqd_deviations / (self._count_data - 1)

    def add_sample(self, value, timestamp, attachments):
        """Adding a sample to Distribution Aggregation Data"""
        self._count_data += 1
        bucket = self.increment_bucket_count(value)

        if attachments is not None and self.exemplars is not None:
            self.exemplars[bucket] = Exemplar(value, timestamp, attachments)
        if self.count_data == 1:
            self._mean_data = value
            return

        old_mean = self._mean_data
        self._mean_data = self._mean_data + (
            (value - self._mean_data) / self._count_data)
        self._sum_of_sqd_deviations = self._sum_of_sqd_deviations + (
            (value - old_mean) * (value - self._mean_data))

    def increment_bucket_count(self, value):
        """Increment the bucket count based on a given value from the user"""
        if len(self._bounds) == 0:
            self._counts_per_bucket[0] += 1
            return 0

        for ii, bb in enumerate(self._bounds):
            if value < bb:
                self._counts_per_bucket[ii] += 1
                return ii
        else:
            last_bucket_index = len(self._bounds)
            self._counts_per_bucket[last_bucket_index] += 1
            return last_bucket_index

    def to_point(self, timestamp):
        """Get a Point conversion of this aggregation.

        This method creates a :class: `opencensus.metrics.export.point.Point`
        with a :class: `opencensus.metrics.export.value.ValueDistribution`
        value, and creates buckets and exemplars for that distribution from the
        appropriate classes in the `metrics` package. If the distribution
        doesn't have a histogram (i.e. `bounds` is empty) the converted point's
        `buckets` attribute will be null.

        :type timestamp: :class: `datetime.datetime`
        :param timestamp: The time to report the point as having been recorded.

        :rtype: :class: `opencensus.metrics.export.point.Point`
        :return: a :class: `opencensus.metrics.export.value.ValueDistribution`
        -valued Point.
        """
        if self.bounds:
            bucket_options = value.BucketOptions(value.Explicit(self.bounds))
            buckets = [None] * len(self.counts_per_bucket)
            for ii, count in enumerate(self.counts_per_bucket):
                stat_ex = self.exemplars.get(ii) if self.exemplars else None
                if stat_ex is not None:
                    metric_ex = value.Exemplar(stat_ex.value,
                                               stat_ex.timestamp,
                                               copy.copy(stat_ex.attachments))
                    buckets[ii] = value.Bucket(count, metric_ex)
                else:
                    buckets[ii] = value.Bucket(count)

        else:
            bucket_options = value.BucketOptions()
            buckets = None
        return point.Point(
            value.ValueDistribution(
                count=self.count_data,
                sum_=self.sum,
                sum_of_squared_deviation=self.sum_of_sqd_deviations,
                bucket_options=bucket_options,
                buckets=buckets
            ),
            timestamp
        )


class LastValueAggregationData(object):
    """
    LastValue Aggregation Data is the value of aggregated data

    :type value_type: class that is either
        :class:`opencensus.metrics.export.value.ValueDouble` or
        :class:`opencensus.metrics.export.value.ValueLong`
    :param value_type: the type of value to be used when creating a point
    :type value: long
    :param value: represents the initial value

    """

    def __init__(self, value_type, value):
        self._value_type = value_type
        self._value = value

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    def add_sample(self, value, timestamp=None, attachments=None):
        """Adds a sample to the current
        LastValue Aggregation Data and overwrite
        the current recorded value"""
        self._value = value

    @property
    def value(self):
        """The current value recorded"""
        return self._value

    @property
    def value_type(self):
        """The value type to use when creating the point"""
        return self._value_type

    def to_point(self, timestamp):
        """Get a Point conversion of this aggregation.

        :type timestamp: :class: `datetime.datetime`
        :param timestamp: The time to report the point as having been recorded.

        :rtype: :class: `opencensus.metrics.export.point.Point`
        :return: a Point with value of type `_value_type`.
        """
        return point.Point(self._value_type(self.value), timestamp)


class Exemplar(object):
    """ Exemplar represents an example point that may be used to annotate
        aggregated distribution values, associated with a histogram bucket.

        :type value: double
        :param value: value of the Exemplar point.

        :type timestamp: time
        :param timestamp: the time that this Exemplar's value was recorded.

        :type attachments: dict
        :param attachments: the contextual information about the example value.
    """

    def __init__(self, value, timestamp, attachments):
        self._value = value

        self._timestamp = timestamp

        if attachments is None:
            raise TypeError('attachments should not be empty')

        for key, value in attachments.items():
            if key is None or not isinstance(key, str):
                raise TypeError('attachment key should not be '
                                'empty and should be a string')
            if value is None or not isinstance(value, str):
                raise TypeError('attachment value should not be '
                                'empty and should be a string')
        self._attachments = attachments

    @property
    def value(self):
        """The current value of the Exemplar point"""
        return self._value

    @property
    def timestamp(self):
        """The time that this Exemplar's value was recorded"""
        return self._timestamp

    @property
    def attachments(self):
        """The contextual information about the example value"""
        return self._attachments



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common.transports import base
from opencensus.trace import execution_context


class SyncTransport(base.Transport):
    def __init__(self, exporter):
        self.exporter = exporter

    def export(self, datas):
        # Used to suppress tracking of requests in export
        execution_context.set_is_exporter(True)
        self.exporter.emit(datas)
        # Reset the context
        execution_context.set_is_exporter(False)



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from opencensus.common.monitored_resource import gcp_metadata_config

# Env var that signals that we're in a kubernetes container
_KUBERNETES_SERVICE_HOST = 'KUBERNETES_SERVICE_HOST'

# Name of the cluster the container is running in
CLUSTER_NAME_KEY = 'k8s.io/cluster/name'

# ID of the instance the container is running on
NAMESPACE_NAME_KEY = 'k8s.io/namespace/name'

# Container pod ID
POD_NAME_KEY = 'k8s.io/pod/name'

# Container name
CONTAINER_NAME_KEY = 'k8s.io/container/name'

# Attributes set from environment variables
_K8S_ENV_ATTRIBUTES = {
    CONTAINER_NAME_KEY: 'CONTAINER_NAME',
    NAMESPACE_NAME_KEY: 'NAMESPACE',
    POD_NAME_KEY: 'HOSTNAME'
}


def is_k8s_environment():
    """Whether the environment is a kubernetes container.

    The KUBERNETES_SERVICE_HOST environment variable must be set.
    """
    return _KUBERNETES_SERVICE_HOST in os.environ


def get_k8s_metadata():
    """Get kubernetes container metadata, as on GCP GKE."""
    k8s_metadata = {}

    gcp_cluster = (gcp_metadata_config.GcpMetadataConfig
                   .get_attribute(gcp_metadata_config.CLUSTER_NAME_KEY))
    if gcp_cluster is not None:
        k8s_metadata[CLUSTER_NAME_KEY] = gcp_cluster

    for attribute_key, attribute_env in _K8S_ENV_ATTRIBUTES.items():
        attribute_value = os.environ.get(attribute_env)
        if attribute_value is not None:
            k8s_metadata[attribute_key] = attribute_value

    return k8s_metadata



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.trace import execution_context, print_exporter, samplers
from opencensus.trace.propagation import trace_context_http_header_format
from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracers import context_tracer, noop_tracer


class Tracer(object):
    """The Tracer is for tracing a request for web applications.

    :type span_context: :class:`~opencensus.trace.span_context.SpanContext`
    :param span_context: SpanContext encapsulates the current context within
                         the request's trace.

    :type sampler: :class:`~opencensus.trace.samplers.base.Sampler`
    :param sampler: Instances of Sampler objects. Defaults to
                    :class:`.ProbabilitySampler`. Other options include
                    :class:`.AlwaysOnSampler` and :class:`.AlwaysOffSampler`.

    :type exporter: :class:`~opencensus.trace.base_exporter.exporter`
    :param exporter: Instances of exporter objects. Default to
                     :class:`.Printexporter`. The rest options are
                     :class:`.Fileexporter`, :class:`.Printexporter`,
                     :class:`.Loggingexporter`, :class:`.Zipkinexporter`,
                     :class:`.GoogleCloudexporter`
    """
    def __init__(
            self,
            span_context=None,
            sampler=None,
            exporter=None,
            propagator=None):
        if span_context is None:
            span_context = SpanContext()

        if sampler is None:
            sampler = samplers.ProbabilitySampler()

        if exporter is None:
            exporter = print_exporter.PrintExporter()

        if propagator is None:
            propagator = \
                trace_context_http_header_format.TraceContextPropagator()

        self.span_context = span_context
        self.sampler = sampler
        self.exporter = exporter
        self.propagator = propagator
        self.tracer = self.get_tracer()
        self.store_tracer()

    def should_sample(self):
        """Determine whether to sample this request or not.
        If the context enables tracing, return True.
        Else follow the decision of the sampler.

        :rtype: bool
        :returns: Whether to trace the request or not.
        """
        return self.sampler.should_sample(self.span_context)

    def get_tracer(self):
        """Return a tracer according to the sampling decision."""
        sampled = self.should_sample()

        if sampled:
            self.span_context.trace_options.set_enabled(True)
            return context_tracer.ContextTracer(
                exporter=self.exporter,
                span_context=self.span_context)
        return noop_tracer.NoopTracer()

    def store_tracer(self):
        """Add the current tracer to thread_local"""
        execution_context.set_opencensus_tracer(self)

    def finish(self):
        """End all spans."""
        self.tracer.finish()

    def span(self, name='span'):
        """Create a new span with the trace using the context information.

        :type name: str
        :param name: The name of the span.

        :rtype: :class:`~opencensus.trace.span.Span`
        :returns: The Span object.
        """
        return self.tracer.span(name)

    def start_span(self, name='span'):
        return self.tracer.start_span(name)

    def end_span(self):
        """End a span. Update the span_id in SpanContext to the current span's
        parent span id; Update the current span; Send the span to exporter.
        """
        self.tracer.end_span()

    def current_span(self):
        """Return the current span."""
        return self.tracer.current_span()

    def add_attribute_to_current_span(self, attribute_key, attribute_value):
        """Add attribute to current span.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        self.tracer.add_attribute_to_current_span(
            attribute_key, attribute_value)

    def trace_decorator(self):
        """Decorator to trace a function."""

        def decorator(func):

            def wrapper(*args, **kwargs):
                self.tracer.start_span(name=func.__name__)
                return_value = func(*args, **kwargs)
                self.tracer.end_span()
                return return_value

            return wrapper

        return decorator



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.tags.validation import is_valid_tag_value

_TAG_VALUE_ERROR = \
    'tag value must not be longer than 255 characters ' \
    'and of ascii values between 32 - 126'


class TagValue(str):
    """The value of a tag"""

    def __new__(cls, value):
        """Create and return a new tag value

        :type value: str
        :param value: A string representing the value of a key in a tag
        :return: TagValue
        """
        if not isinstance(value, cls):
            if not is_valid_tag_value(value):
                raise ValueError(_TAG_VALUE_ERROR)
        return super(TagValue, cls).__new__(cls, value)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common import resource
from opencensus.common.monitored_resource import (
    aws_identity_doc_utils,
    gcp_metadata_config,
    k8s_utils,
)

# Supported environments (resource types)
_GCE_INSTANCE = "gce_instance"
_K8S_CONTAINER = "k8s_container"
_AWS_EC2_INSTANCE = "aws_ec2_instance"


def is_gce_environment():
    """Whether the environment is a virtual machine on GCE."""
    return gcp_metadata_config.GcpMetadataConfig.is_running_on_gcp()


def is_aws_environment():
    """Whether the environment is a virtual machine instance on EC2."""
    return aws_identity_doc_utils.AwsIdentityDocumentUtils.is_running_on_aws()


def get_instance():
    """Get a resource based on the application environment.

    Returns a `Resource` configured for the current environment, or None if the
    environment is unknown or unsupported.

    :rtype: :class:`opencensus.common.resource.Resource` or None
    :return: A `Resource` configured for the current environment.
    """
    resources = []
    env_resource = resource.get_from_env()
    if env_resource is not None:
        resources.append(env_resource)

    if k8s_utils.is_k8s_environment():
        resources.append(resource.Resource(
            _K8S_CONTAINER, k8s_utils.get_k8s_metadata()))

    if is_gce_environment():
        resources.append(resource.Resource(
            _GCE_INSTANCE,
            gcp_metadata_config.GcpMetadataConfig().get_gce_metadata()))
    elif is_aws_environment():
        resources.append(resource.Resource(
            _AWS_EC2_INSTANCE,
            (aws_identity_doc_utils.AwsIdentityDocumentUtils()
             .get_aws_metadata())))

    if not resources:
        return None
    return resource.merge_resources(resources)



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from opencensus.metrics.export import gauge, metric_descriptor


class CumulativePointLong(gauge.GaugePointLong):
    """A `GaugePointLong` that cannot decrease."""

    def _set(self, val):
        if not isinstance(val, six.integer_types):
            raise ValueError("CumulativePointLong only supports integer types")
        if val > self.get_value():
            super(CumulativePointLong, self)._set(val)

    def add(self, val):
        """Add `val` to the current value if it's positive.

        Return without adding if `val` is not positive.

        :type val: int
        :param val: Value to add.
        """
        if not isinstance(val, six.integer_types):
            raise ValueError("CumulativePointLong only supports integer types")
        if val > 0:
            super(CumulativePointLong, self).add(val)


class CumulativePointDouble(gauge.GaugePointDouble):
    """A `GaugePointDouble` that cannot decrease."""

    def _set(self, val):
        if val > self.get_value():
            super(CumulativePointDouble, self)._set(val)

    def add(self, val):
        """Add `val` to the current value if it's positive.

        Return without adding if `val` is not positive.

        :type val: float
        :param val: Value to add.
        """
        if val > 0:
            super(CumulativePointDouble, self).add(val)


class LongCumulativeMixin(object):
    """Type mixin for long-valued cumulative measures."""
    descriptor_type = metric_descriptor.MetricDescriptorType.CUMULATIVE_INT64
    point_type = CumulativePointLong


class DoubleCumulativeMixin(object):
    """Type mixin for float-valued cumulative measures."""
    descriptor_type = metric_descriptor.MetricDescriptorType.CUMULATIVE_DOUBLE
    point_type = CumulativePointDouble


class LongCumulative(LongCumulativeMixin, gauge.Gauge):
    """Records cumulative int-valued measurements."""


class DoubleCumulative(DoubleCumulativeMixin, gauge.Gauge):
    """Records cumulative float-valued measurements."""


class DerivedLongCumulative(LongCumulativeMixin, gauge.DerivedGauge):
    """Records derived cumulative int-valued measurements."""


class DerivedDoubleCumulative(DoubleCumulativeMixin, gauge.DerivedGauge):
    """Records derived cumulative float-valued measurements."""



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

_INTEGRATIONS_BIT_MASK = 0
_INTEGRATIONS_LOCK = threading.Lock()


class _Integrations:
    NONE = 0
    DJANGO = 1
    FLASK = 2
    GOOGLE_CLOUD = 4
    HTTP_LIB = 8
    LOGGING = 16
    MYSQL = 32
    POSTGRESQL = 64
    PYMONGO = 128
    PYMYSQL = 256
    PYRAMID = 512
    REQUESTS = 1024
    SQLALCHEMY = 2056
    HTTPX = 4096
    FASTAPI = 8192


def get_integrations():
    return _INTEGRATIONS_BIT_MASK


def add_integration(integration):
    with _INTEGRATIONS_LOCK:
        global _INTEGRATIONS_BIT_MASK  # pylint: disable=global-statement
        _INTEGRATIONS_BIT_MASK |= integration


def remove_intregration(integration):
    with _INTEGRATIONS_LOCK:
        global _INTEGRATIONS_BIT_MASK  # pylint: disable=global-statement
        _INTEGRATIONS_BIT_MASK &= ~integration



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common import utils


class Type(object):
    """
    Indicates whether the message was sent or received.

    Attributes:
      TYPE_UNSPECIFIED (int): Unknown event type.
      SENT (int): Indicates a sent message.
      RECEIVED (int): Indicates a received message.
    """
    TYPE_UNSPECIFIED = 0
    SENT = 1
    RECEIVED = 2


class Annotation(object):
    """Text annotation with a set of attributes.

    :type timestamp: :class:`~datetime.datetime`
    :param timestamp: The timestamp indicating the time the event occurred.

    :type description: str
    :param description: A user-supplied message describing the event.
                        The maximum length for the description is 256 bytes.

    :type attributes: :class:`~opencensus.trace.attributes.Attributes`
    :param attributes: A set of attributes on the annotation.
                       You can have up to 4 attributes per Annotation.
    """
    def __init__(self, timestamp, description, attributes=None):
        self.timestamp = utils.to_iso_str(timestamp)
        self.description = description
        self.attributes = attributes

    def format_annotation_json(self):
        annotation_json = {}
        annotation_json['description'] = utils.get_truncatable_str(
            self.description)

        if self.attributes is not None:
            annotation_json['attributes'] = self.attributes.\
                format_attributes_json()

        return annotation_json


class MessageEvent(object):
    """An event describing a message sent/received between Spans.

    :type timestamp: :class:`~datetime.datetime`
    :param timestamp: The timestamp indicating the time the event occurred.

    :type type: Enum of :class: `~opencensus.trace.time_event.Type`
    :param type: Indicates whether the message was sent or received.

    :type id: str (int64 format)
    :param id: An identifier for the MessageEvent's message that can be used
               to match SENT and RECEIVED MessageEvents. It is recommended to
               be unique within a Span.

    :type uncompressed_size_bytes: str (int64 format)
    :param uncompressed_size_bytes: The number of uncompressed bytes sent or
                                    received.

    :type compressed_size_bytes: str (int64 format)
    :param compressed_size_bytes: The number of compressed bytes sent or
                                  received. If missing assumed to be the same
                                  size as uncompressed.

    """
    def __init__(self, timestamp, id, type=None, uncompressed_size_bytes=None,
                 compressed_size_bytes=None):
        self.timestamp = utils.to_iso_str(timestamp)

        if type is None:
            type = Type.TYPE_UNSPECIFIED

        if compressed_size_bytes is None and \
                uncompressed_size_bytes is not None:
            compressed_size_bytes = uncompressed_size_bytes

        self.id = id
        self.type = type
        self.uncompressed_size_bytes = uncompressed_size_bytes
        self.compressed_size_bytes = compressed_size_bytes

    def format_message_event_json(self):
        message_event_json = {}

        message_event_json['id'] = self.id
        message_event_json['type'] = self.type

        if self.uncompressed_size_bytes is not None:
            message_event_json[
                'uncompressed_size_bytes'] = self.uncompressed_size_bytes

        if self.compressed_size_bytes is not None:
            message_event_json[
                'compressed_size_bytes'] = self.compressed_size_bytes

        return message_event_json
